
# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any

from neuro_san.test.interfaces.assert_forwarder import AssertForwarder


class NullAssertForwarder(AssertForwarder):
    """
    Implementation of the AssertForwarder interface that does nothing for each assert type.
    This is useful as a superclass for other AssertForwarders that only want to focus
    on a few specific asserts.
    """

    # pylint: disable=invalid-name
    def assertEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is equal to the second

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertNotEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is not equal to the second

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertTrue(self, expr: Any, msg: str = None):
        """
        Assert that the expression is true

        :param expr: Expression to test
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertFalse(self, expr: Any, msg: str = None):
        """
        Assert that the expression is false

        :param expr: Expression to test
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertIs(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first and second are the same object

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertIsNot(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first and second are not the same object

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertIsNone(self, expr: Any, msg: str = None):
        """
        Assert that the expression is None

        :param expr: Expression to test
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertIsNotNone(self, expr: Any, msg: str = None):
        """
        Assert that the expression is not None

        :param expr: Expression to test
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertIn(self, member: Any, container: Any, msg: str = None):
        """
        Assert that the member is in the container

        :param member: Member comparison element
        :param container: Container comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertNotIn(self, member: Any, container: Any, msg: str = None):
        """
        Assert that the member is not in the container

        :param member: Member comparison element
        :param container: Container comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertIsInstance(self, obj: Any, cls: Any, msg: str = None):
        """
        Assert that the obj is an instance of the cls

        :param obj: object instance comparison element
        :param cls: Class comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertNotIsInstance(self, obj: Any, cls: Any, msg: str = None):
        """
        Assert that the obj is not an instance of the cls

        :param obj: object instance comparison element
        :param cls: Class comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertGreater(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is greater than the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertGreaterEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is greater than or equal to the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertLess(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is less than the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertLessEqual(self, first: Any, second: Any, msg: str = None):
        """
        Assert that the first is less than or equal to the second.

        :param first: First comparison element
        :param second: Second comparison element
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertGist(self, gist: bool, acceptance_criteria: str, text_sample: str, msg: str = None):
        """
        Assert that the gist is true

        :param gist: Pass/Fail value of the gist expected to be True
        :param acceptance_criteria: The value to verify against
        :param text_sample: The value appearing in the test sample
        :param msg: optional string message
        """
        # Do nothing

    # pylint: disable=invalid-name
    def assertNotGist(self, gist: bool, acceptance_criteria: str, text_sample: str, msg: str = None):
        """
        Assert that the gist is true

        :param gist: Pass/Fail value of the gist expected to be False
        :param acceptance_criteria: The value to verify against
        :param text_sample: The value appearing in the test sample
        :param msg: optional string message
        """
        # Do nothing




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.interfaces.concierge_session import ConciergeSession
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage


class DirectConciergeSession(ConciergeSession):
    """
    Service-agnostic guts for a ConciergeSession.
    This could be used by an HTTP service.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self,
                 network_storage: AgentNetworkStorage,
                 metadata: Dict[str, Any] = None,
                 security_cfg: Dict[str, Any] = None):
        """
        Constructor

        :param network_storage: A AgentNetworkStorage instance which keeps all
                                the AgentNetwork instances.
        :param metadata: A dictionary of request metadata to be forwarded
                        to subsequent yet-to-be-made requests.
        :param security_cfg: A dictionary of parameters used to
                        secure the TLS and the authentication of the HTTP
                        connection.  Supplying this implies use of a secure
                        HTTP connection.  If None, uses insecure connection.
        """
        self.network_storage: AgentNetworkStorage = network_storage
        # These aren't used yet
        self._metadata: Dict[str, Any] = metadata
        self._security_cfg: Dict[str, Any] = security_cfg

    def list(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConciergeRequest
                    protobuf structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConciergeResponse
                    protobuf structure. Has the following keys:
                "agents" - the sequence of dictionaries describing available agents
        """
        agents_names: List[str] = self.network_storage.get_agent_names()
        empty_list: List[str] = []

        agents_list: List[Dict[str, Any]] = []
        for agent_name in agents_names:

            # Get the spec for the agent network
            provider: AgentNetworkProvider = self.network_storage.get_agent_network_provider(agent_name)
            agent_network: AgentNetwork = provider.get_agent_network()
            agent_spec: Dict[str, Any] = agent_network.get_config()
            extractor = DictionaryExtractor(agent_spec)

            # It's concievable we could get the description from the front man's function.
            # We haven't done that yet, though, so deferring until a hew and cry emerges.
            description: str = extractor.get("metadata.description", "")
            tags: List[str] = extractor.get("metadata.tags", empty_list)

            # Construct an AgentInfo entry
            agent_info: Dict[str, Any] = {
                "agent_name": agent_name,
                "description": description,
                "tags": tags,
            }
            agents_list.append(agent_info)

        response: Dict[str, Any] = {
            "agents": agents_list
        }
        return response




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from numbers import Number
from typing import Any
from typing import Dict
from typing import List

from copy import copy

from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.message_processing.message_processor import MessageProcessor


class ChatHistoryMessageProcessor(MessageProcessor):
    """
    MessageProcessor implementation for processing a single message
    in a chat history.
    """

    def __init__(self, max_message_history: int = None):
        """
        Constructor

        :param max_message_history: The maximum number of messages to preserve
                in the message history, not including the instructions.
                The default value of None implies there is no maximum.
                Non-positive numbers revert to no-maximum behavior.
        """
        self.max_message_history: int = max_message_history
        if self.max_message_history is not None:
            if not isinstance(self.max_message_history, Number):
                # If we don't have a number we don't have a max.
                self.max_message_history = None
            else:
                # Be sure we are dealing with an integer
                self.max_message_history = int(self.max_message_history)

        self.message_history: List[Dict[str, Any]] = []
        self.saw_first_system: bool = False

    def get_message_history(self) -> List[Dict[str, Any]]:
        """
        :return: The filtered message history
        """
        return self.message_history

    def process_messages(self, chat_message_dicts: List[Dict[str, Any]]):
        """
        Convenience method for processing lists of messages.
        :param chat_message_dicts: The messages to process.
        """
        super().process_messages(chat_message_dicts)

        # See if we need to curtail chat history at all.
        if self.max_message_history is None or self.max_message_history <= 1:
            # Nothing to do.
            return

        # Save the first item in the list. This is the redacted placeholder for instructions
        instructions: Dict[str, Any] = self.message_history.pop(0)

        # Preserve the most recent n elements in the message history
        self.message_history = self.message_history[-self.max_message_history:]

        # Prepend the instructions to the list
        self.message_history.insert(0, instructions)

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        """
        if message_type not in (ChatMessageType.HUMAN, ChatMessageType.SYSTEM, ChatMessageType.AI):
            # Don't send any messages over the wire that won't be re-ingestable.
            return

        transformed_message_dict: Dict[str, Any] = chat_message_dict
        if not self.saw_first_system and message_type == ChatMessageType.SYSTEM:
            # Redact the first SYSTEM message we see. This has the front-man prompt in it,
            # and when read in, we replace it with what the agent has anyway to prevent
            # a prompting takeover.
            transformed_message_dict = self.redact_instructions(chat_message_dict)
            self.saw_first_system = True
        else:
            # Transform the message with properly escaped text.
            transformed_message_dict = self.escape_message(chat_message_dict)

        if transformed_message_dict is not None:
            self.message_history.append(transformed_message_dict)

    def redact_instructions(self, chat_message_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Redacts the text instructions of the given system message.
        These will get replaced by the agent's instructions anyway to preserve
        server-side intent.
        """
        redacted: Dict[str, Any] = copy(chat_message_dict)
        redacted["text"] = "<redacted>"
        return redacted

    def escape_message(self, chat_message_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Prepare a message such that it can be re-ingested by the system nicely.
        This means properly escaping any text that is sent.
        """
        transformed: Dict[str, Any] = copy(chat_message_dict)
        text: str = transformed.get("text")

        if text is None:
            return None

        # Braces are a problem for chat history being read back into the system
        # if they are not properly escaped.

        # First replace any pre-escaped braces with normal braces
        text = text.replace("{{", "{")
        text = text.replace("}}", "}")

        # Now replace normal braces with escaped braces.
        # Idea is to catch everything pre-escaped or not
        text = text.replace("{", "{{")
        text = text.replace("}", "}}")

        # JSON spec does not allow control characters in strings and newlines in particular
        # can be a problem for http clients that expect one full JSON message per line.
        # Replace any lurking newlines with the 2 raw characters \ and n.
        # DEF - for the future.

        transformed["text"] = text
        return transformed




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from neuro_san.api.grpc import concierge_pb2 as neuro__san_dot_api_dot_grpc_dot_concierge__pb2

GRPC_GENERATED_VERSION = '1.78.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in neuro_san/api/grpc/concierge_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class ConciergeServiceStub(object):
    """The service provides system-wide information about resources available
    in current instance of neuro-san.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.List = channel.unary_unary(
                '/dev.cognizant_ai.neuro_san.api.grpc.concierge.ConciergeService/List',
                request_serializer=neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeRequest.SerializeToString,
                response_deserializer=neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeResponse.FromString,
                _registered_method=True)


class ConciergeServiceServicer(object):
    """The service provides system-wide information about resources available
    in current instance of neuro-san.
    """

    def List(self, request, context):
        """Called when a client needs the information about available agents.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ConciergeServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'List': grpc.unary_unary_rpc_method_handler(
                    servicer.List,
                    request_deserializer=neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeRequest.FromString,
                    response_serializer=neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'dev.cognizant_ai.neuro_san.api.grpc.concierge.ConciergeService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('dev.cognizant_ai.neuro_san.api.grpc.concierge.ConciergeService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class ConciergeService(object):
    """The service provides system-wide information about resources available
    in current instance of neuro-san.
    """

    @staticmethod
    def List(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/dev.cognizant_ai.neuro_san.api.grpc.concierge.ConciergeService/List',
            neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeRequest.SerializeToString,
            neuro__san_dot_api_dot_grpc_dot_concierge__pb2.ConciergeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from neuro_san.internals.errors.json_error_formatter import JsonErrorFormatter
from neuro_san.internals.errors.string_error_formatter import StringErrorFormatter
from neuro_san.internals.interfaces.error_formatter import ErrorFormatter


class ErrorFormatterFactory:
    """
    Factory class to create an appropriate ErrorFormatter
    """

    @staticmethod
    def create_formatter(name: str = "string") -> ErrorFormatter:
        """
        Creates an ErrorFormatter given the name.

        :param name: The name of the error formatter to use
        :return: An ErrorFormatter instance.
        """

        # Default
        formatter: ErrorFormatter = StringErrorFormatter()
        if name is None:
            return formatter

        if name.lower() == "json":
            formatter = JsonErrorFormatter()

        # When the need arises, we could conceivably add class name lookup
        # for error formatters here, not unlike the way we do for coded tools.

        return formatter




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict


class EventLoopLogger:
    """
    Interface for a logger used in event loop request processing.
    Each logger call has request-specific metadata,
    which should be presented in the logger output.
    """

    def info(self, metadata: Dict[str, Any], msg: str, *args):
        """
        "Info" logging method.
        Prepare logger filter with request-specific metadata
        and delegate logging to underlying standard Logger.
        """
        raise NotImplementedError

    def warning(self, metadata: Dict[str, Any], msg: str, *args):
        """
        "Warning" logging method.
        Prepare logger filter with request-specific metadata
        and delegate logging to underlying standard Logger.
        """
        raise NotImplementedError

    def debug(self, metadata: Dict[str, Any], msg: str, *args):
        """
        "Debug" logging method.
        Prepare logger filter with request-specific metadata
        and delegate logging to underlying standard Logger.
        """
        raise NotImplementedError

    def error(self, metadata: Dict[str, Any], msg: str, *args):
        """
        "Error" logging method.
        Prepare logger filter with request-specific metadata
        and delegate logging to underlying standard Logger.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import List

import http
import os

from importlib.metadata import version as library_version
from importlib.metadata import PackageNotFoundError

from tornado.web import RequestHandler
from neuro_san.service.http.logging.http_logger import HttpLogger
from neuro_san.service.utils.server_status import ServerStatus


class HealthCheckHandler(RequestHandler):
    """
    Handler class for API endpoint health check.
    """

    # pylint: disable=attribute-defined-outside-init
    def initialize(self,
                   forwarded_request_metadata: List[str],
                   server_status: ServerStatus,
                   op: str,
                   logging_config: Dict[str, Any]):
        """
        This method is called by Tornado framework to allow
        injecting service-specific data into local handler context.
        Here we use it to inject CORS headers if so configured.
        :param forwarded_request_metadata: list of client metadata keys;
        :param server_status: current server status to query;
        :param op: requested healthcheck operation:
                   "ready" for /readyz query
                   "live" for /livez query
        :param logging_config: logging configuration dictionary
        """
        self.logger = HttpLogger(forwarded_request_metadata, logging_config)
        if op == "ready":
            self.status = server_status.is_server_ready()
        else:
            self.status = server_status.is_server_live()
        self.server_name = server_status.get_server_name()

        if os.environ.get("AGENT_ALLOW_CORS_HEADERS") is not None:
            self.set_header("Access-Control-Allow-Origin", "*")
            self.set_header("Access-Control-Allow-Methods", "GET, OPTIONS")
            self.set_header("Access-Control-Allow-Headers", "Content-Type, Transfer-Encoding")

    async def get(self):
        """
        Implementation of GET request handler for API health check.
        """

        try:
            if self.status:
                versions: Dict[str, Any] = self.determine_versions()
                result_dict: Dict[str, Any] = {
                    "service": self.server_name,
                    "status": "ok",
                    "versions": versions
                }
                self.set_header("Content-Type", "application/json")
                self.write(result_dict)
            else:
                # Set "service unavailable" status
                self.set_status(503)
                self.write({"error": "Service Unavailable"})
        except Exception:  # pylint: disable=broad-exception-caught
            # Handle unexpected errors
            self.set_status(500)
            self.write({"error": "Internal server error"})
        finally:
            self.finish()

    def get_metadata(self) -> Dict[str, Any]:
        """
        Get request metadata
        """
        return {}

    def data_received(self, chunk):
        """
        Method overrides abstract method of RequestHandler
        with no-op implementation.
        """
        return

    async def options(self, *_args, **_kwargs):
        """
        Handles OPTIONS requests for CORS support
        """
        # No body needed. Just return a 204 No Content
        self.set_status(http.HTTPStatus.NO_CONTENT)
        self.finish()

    def determine_versions(self) -> Dict[str, Any]:
        """
        Allow for a list of libraries to report versioning.
        We have a static list that we always support, but also look at the
        AGENT_VERSION_LIBS env var (if set) to potentially report more.
        :return: A dictionary whose keys are libraries and whose values are the
                 versions installed for those libraries.
        """
        versions: Dict[str, Any] = {}

        # Start with a static list of libs we always support
        versioned_libs: List[str] = ["neuro-san"]

        # Try to incorporate libs from the env var.
        env_var_libs: str = os.environ.get("AGENT_VERSION_LIBS")
        if env_var_libs is not None:
            for env_var_lib in env_var_libs.split(" "):
                if env_var_lib is None:
                    # Skip anything with nothing in it
                    continue

                env_var_lib = env_var_lib.strip()
                if len(env_var_lib) == 0:
                    # Skip anything with nothing in it
                    continue

                if ":" not in env_var_lib:
                    # Look for the library in the versioning loop below
                    versioned_libs.append(env_var_lib)
                    continue

                # We have a version coming in from the env var
                version: str = env_var_lib.split(":")[1]
                env_var_lib = env_var_lib.split(":")[0]
                versions[env_var_lib] = version

        for lib in versioned_libs:
            # Did we already find a version from the env var?
            version: str = versions.get(lib)
            if version is None:
                try:
                    # Try getting the pip-installed version from the system
                    version = str(library_version(lib))
                except PackageNotFoundError:
                    # Don't know what to do, so report a shrug.
                    version = "unknown"
                versions[lib] = version

        return versions




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Union


class MetadataUtil:
    """
    Utilities to deal with metadata dictionaries.
    """

    @staticmethod
    def minimize_metadata(request_metadata: Dict[str, Any],
                          keys: Union[str, List[str]]) -> Dict[str, Any]:
        """
        :param request_metadata: The raw request metadata dictionary that could easily contain
                    more keys than we want to send to the UsageLogger.
        :param keys: The keys we want to send to the UsageLogger
                This can either be a single string with space-delimited keys, or a list of keys
        :return: A minimized dictionary that only sends the keys we need to the UsageLogger.
                The idea is that this prevents the UsageLogger from getting potentially
                sensitive information it shouldn't really have.

                If the requested keys in the metadata are not there, they will also not appear
                in the returned minimized dictionary.
        """
        minimized: Dict[str, Any] = {}

        if request_metadata is None:
            # No request data, nothing to fill
            return minimized

        # If there are no keys, there is nothing to fill.
        if not keys:
            return minimized

        keys_list: List[str] = []

        # Check if keys is a string or a list
        if isinstance(keys, str):
            keys_list = keys.split(" ")
        elif isinstance(keys, List):
            keys_list = keys

        for key in keys_list:

            if not key:
                # Skip any empty key split from the list. Allows for multi-spaces.
                continue

            value: str = request_metadata.get(key)
            if value:
                minimized[key] = value

        return minimized




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider


class FixedAgentNetworkProvider(AgentNetworkProvider):
    """
    Class providing fixed immutable AgentNetwork for a given agent in the service scope.
    """
    def __init__(self, agent_network: AgentNetwork):
        """
        Constructor.
        :param agent_network: AgentNetwork instance to be returned by this provider.
        """
        self.agent_network: AgentNetwork = agent_network

    def get_agent_network(self) -> AgentNetwork:
        """
        :return: Current AgentNetwork instance for specific agent name.
        """
        return self.agent_network




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from neuro_san.interfaces.agent_session import AgentSession


class ConciergeSession:
    """
    Interface for a Concierge session.
    """

    # Default port for the Concierge HTTP Service
    # This port number will also be mentioned in its Dockerfile
    DEFAULT_HTTP_PORT: int = AgentSession.DEFAULT_HTTP_PORT

    # Default port for the Concierge Service
    # This port number will also be mentioned in its Dockerfile
    DEFAULT_PORT: int = DEFAULT_HTTP_PORT

    def list(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConciergeRequest
                    protobuf structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConciergeResponse
                    protobuf structure. Has the following keys:
                "agents" - the sequence of dictionaries describing available agents
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

import argparse
import json
import os
import shutil
import sys

from pyparsing.exceptions import ParseException
from pyparsing.exceptions import ParseSyntaxException

from neuro_san.internals.interfaces.dictionary_validator import DictionaryValidator
from neuro_san.internals.graph.persistence.agent_network_restorer import AgentNetworkRestorer
from neuro_san.internals.validation.network.manifest_network_validator import ManifestNetworkValidator


class HoconValidatorCli:
    """
    Command-line tool for validating HOCON agent network configuration files.

    This script validates HOCON files against neuro-san's agent network validation rules,
    checking for issues such as:
    - Missing or unreachable agents
    - Invalid tool names
    - Empty instructions
    - Invalid URL references

    Usage:
        python -m neuro_san.client.hocon_validator_cli path/to/agent.hocon
        python -m neuro_san.client.hocon_validator_cli path/to/agent.hocon --verbose
    """

    def __init__(self):
        """
        Constructor
        """
        self.args = None

    def main(self) -> int:
        """
        Main entry point for the HOCON validator CLI.

        :return: Exit code (0 for success, 1 for validation errors, 2 for other errors)
        """
        self.parse_args()

        try:
            config: Dict[str, Any] = self.load_hocon_file(self.args.hocon_file)
        except FileNotFoundError as exception:
            print(f"Error: File not found - {exception}", file=sys.stderr)
            return 2
        except (ParseException, ParseSyntaxException) as exception:
            print(f"Error: Failed to parse HOCON file - {exception}", file=sys.stderr)
            return 2
        except ValueError as exception:
            print(f"Error: {exception}", file=sys.stderr)
            return 2

        validator: DictionaryValidator = self.create_validator()
        errors: List[str] = validator.validate(config)

        if errors:
            print(f"Validation failed with {len(errors)} error(s):\n")
            for i, error in enumerate(errors, 1):
                print(f"  {i}. {error}")
            return 1

        print("Validation passed: No errors found.")
        if self.args.verbose:
            self.print_network_summary(config)
        return 0

    def parse_args(self):
        """
        Parse command line arguments.
        """
        arg_parser = argparse.ArgumentParser(
            description="Validate a HOCON agent network configuration file.",
            formatter_class=argparse.RawDescriptionHelpFormatter,
            epilog="""
Examples:
  python -m neuro_san.client.hocon_validator_cli registries/hello_world.hocon
  python -m neuro_san.client.hocon_validator_cli my_agent.hocon --verbose
            """
        )

        arg_parser.add_argument(
            "hocon_file",
            type=str,
            help="Path to the HOCON file to validate"
        )

        arg_parser.add_argument(
            "--verbose",
            default=False,
            action="store_true",
            help="Print additional information about the agent network"
        )

        arg_parser.add_argument(
            "--external-agents",
            type=str,
            default=None,
            dest="external_agents",
            help="Comma-separated list of valid external agent references (e.g., '/agent1,/agent2')"
        )

        arg_parser.add_argument(
            "--mcp-servers",
            type=str,
            default=None,
            dest="mcp_servers",
            help="Comma-separated list of valid MCP server URLs"
        )

        arg_parser.add_argument(
            "--json-output",
            default=False,
            action="store_true",
            dest="json_output",
            help="Output validation results as JSON"
        )

        # Determine default registry directory from AGENT_MANIFEST_FILE if set
        agent_manifest_file: str = os.environ.get("AGENT_MANIFEST_FILE")
        if agent_manifest_file:
            default_registry_dir: str = os.path.dirname(
                os.path.dirname(os.path.abspath(agent_manifest_file))
            )
        else:
            default_registry_dir: str = os.getcwd()

        arg_parser.add_argument(
            "--registry-dir",
            type=str,
            default=default_registry_dir,
            dest="registry_dir",
            help="Base directory containing the registries folder for resolving HOCON includes. "
                 "Defaults to the parent directory of AGENT_MANIFEST_FILE. "
                 "Use this when your HOCON file has includes like 'include \"registries/...\"'"
        )

        self.args = arg_parser.parse_args()

    def load_hocon_file(self, file_path: str) -> Dict[str, Any]:
        """
        Load and parse a HOCON file.

        If registry_dir is specified, the file is temporarily copied to that directory
        so that HOCON includes (like 'include "registries/..."') can be resolved correctly.

        :param file_path: Path to the HOCON file
        :return: Parsed configuration dictionary
        """
        abs_file_path: str = os.path.abspath(file_path)

        if self.args.registry_dir:
            return self._load_with_registry_dir(abs_file_path, self.args.registry_dir)

        restorer = AgentNetworkRestorer(registry_dir=None)
        agent_network = restorer.restore(file_reference=abs_file_path)
        return agent_network.get_config()

    def _load_with_registry_dir(self, file_path: str, registry_dir: str) -> Dict[str, Any]:
        """
        Load a HOCON file by temporarily copying it to the registry directory.

        This allows HOCON includes to be resolved relative to the registry directory.

        :param file_path: Absolute path to the HOCON file
        :param registry_dir: Directory containing the registries folder
        :return: Parsed configuration dictionary
        """
        abs_registry_dir: str = os.path.abspath(registry_dir)

        if not os.path.isdir(abs_registry_dir):
            raise ValueError(f"Registry directory does not exist: {abs_registry_dir}")

        temp_file: str = None
        try:
            base_name: str = os.path.basename(file_path)
            temp_file = os.path.join(abs_registry_dir, f"_temp_validate_{base_name}")

            shutil.copy(file_path, temp_file)

            restorer = AgentNetworkRestorer(registry_dir=None)
            agent_network = restorer.restore(file_reference=temp_file)
            return agent_network.get_config()
        finally:
            if temp_file and os.path.exists(temp_file):
                os.remove(temp_file)

    def create_validator(self) -> DictionaryValidator:
        """
        Create the validator using ManifestNetworkValidator.

        :return: A DictionaryValidator instance
        """
        external_agents: List[str] = None
        if self.args.external_agents:
            external_agents = [agent.strip() for agent in self.args.external_agents.split(",")]

        mcp_servers: List[str] = None
        if self.args.mcp_servers:
            mcp_servers = [server.strip() for server in self.args.mcp_servers.split(",")]

        return ManifestNetworkValidator(external_agents, mcp_servers)

    def print_network_summary(self, config: Dict[str, Any]):
        """
        Print a summary of the agent network.

        :param config: The agent network configuration
        """
        tools: List[Dict[str, Any]] = config.get("tools", [])

        print("\n--- Agent Network Summary ---")
        print(f"Total agents/tools defined: {len(tools)}")

        if tools:
            print("\nAgents:")
            for tool in tools:
                name: str = tool.get("name", "<unnamed>")
                has_instructions: bool = tool.get("instructions") is not None
                sub_tools: List[str] = tool.get("tools", [])
                tool_type: str = "LLM Agent" if has_instructions else "Coded Tool"
                print(f"  - {name} ({tool_type})")
                if sub_tools:
                    print(f"      Sub-tools: {', '.join(str(t) for t in sub_tools)}")

        metadata: Dict[str, Any] = config.get("metadata", {})
        if metadata:
            print(f"\nMetadata: {json.dumps(metadata, indent=2)}")


if __name__ == "__main__":
    sys.exit(HoconValidatorCli().main())




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from neuro_san.api.grpc import agent_pb2 as neuro__san_dot_api_dot_grpc_dot_agent__pb2

GRPC_GENERATED_VERSION = '1.78.0'
GRPC_VERSION = grpc.__version__
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    raise RuntimeError(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + ' but the generated code in neuro_san/api/grpc/agent_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
    )


class AgentServiceStub(object):
    """The service comprises all the exchanges to the backend in support of a single agent's
    services.  Routing is done by way of agent name on the grpc service hosting the agent,
    so as to keep info about which agents are hosted private (grpc gives the hand when a
    particular agent is unknown.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Function = channel.unary_unary(
                '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/Function',
                request_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionRequest.SerializeToString,
                response_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionResponse.FromString,
                _registered_method=True)
        self.StreamingChat = channel.unary_stream(
                '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/StreamingChat',
                request_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatRequest.SerializeToString,
                response_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatResponse.FromString,
                _registered_method=True)
        self.Connectivity = channel.unary_unary(
                '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/Connectivity',
                request_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityRequest.SerializeToString,
                response_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityResponse.FromString,
                _registered_method=True)


class AgentServiceServicer(object):
    """The service comprises all the exchanges to the backend in support of a single agent's
    services.  Routing is done by way of agent name on the grpc service hosting the agent,
    so as to keep info about which agents are hosted private (grpc gives the hand when a
    particular agent is unknown.
    """

    def Function(self, request, context):
        """Called when a client needs the function description of an agent.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamingChat(self, request, context):
        """Most important semantics of the streaming:
        1) The "answer" to a query of any agent network is in the *last* streamed
        AGENT_FRAMEWORK message.
        2) To RESTfully continue your conversation with the agent network:
        The very last AGENT_FRAMEWORK message before the stream closes will
        have its chat_context field filled in with a structure. You can
        copy this whole-cloth to the chat_context of your next StreamingChat
        request to continue the conversation.
        3) It is important to note that since this is a streaming API, for HTTP clients:
        a) Any single response will always be on the same line. That is,
        responses will not be broken up across multiple lines in an
        HTTP response.
        b) We cannot yet guarantee that there will be only one streamed response
        per HTTP response line. That is, it is possible for more than one
        response *might* come on a single line if they come quickly enough,
        though this is not the empirically observed norm.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Connectivity(self, request, context):
        """Called when a client needs the internal connectivity description of an agent.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AgentServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Function': grpc.unary_unary_rpc_method_handler(
                    servicer.Function,
                    request_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionRequest.FromString,
                    response_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionResponse.SerializeToString,
            ),
            'StreamingChat': grpc.unary_stream_rpc_method_handler(
                    servicer.StreamingChat,
                    request_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatRequest.FromString,
                    response_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatResponse.SerializeToString,
            ),
            'Connectivity': grpc.unary_unary_rpc_method_handler(
                    servicer.Connectivity,
                    request_deserializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityRequest.FromString,
                    response_serializer=neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class AgentService(object):
    """The service comprises all the exchanges to the backend in support of a single agent's
    services.  Routing is done by way of agent name on the grpc service hosting the agent,
    so as to keep info about which agents are hosted private (grpc gives the hand when a
    particular agent is unknown.
    """

    @staticmethod
    def Function(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/Function',
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionRequest.SerializeToString,
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.FunctionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StreamingChat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/StreamingChat',
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatRequest.SerializeToString,
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.ChatResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Connectivity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/dev.cognizant_ai.neuro_san.api.grpc.agent.AgentService/Connectivity',
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityRequest.SerializeToString,
            neuro__san_dot_api_dot_grpc_dot_agent__pb2.ConnectivityResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List

from logging import getLogger
from logging import Logger

from neuro_san.internals.validation.network.abstract_network_validator import AbstractNetworkValidator


class KeywordNetworkValidator(AbstractNetworkValidator):
    """
    AgentNetworkValidator that looks for correct keywords in an agent network
    """

    def __init__(self):
        """
        Constructor
        """
        self.logger: Logger = getLogger(self.__class__.__name__)

    def validate_name_to_spec_dict(self, name_to_spec: Dict[str, Any]) -> List[str]:
        """
        Validate the agent network, specifically in the form of a name -> agent spec dictionary.

        :param name_to_spec: The name -> agent spec dictionary to validate
        :return: List of errors indicating agents and missing keywords
        """
        errors: List[str] = []

        self.logger.info("Validating agent network keywords...")

        # Currently, only required "instructions" for non-function agents.
        for agent_name, agent in name_to_spec.items():
            if agent.get("instructions") == "":
                error_msg = f"{agent_name} 'instructions' cannot be empty."
                errors.append(error_msg)

        # Only warn if there is a problem
        if len(errors) > 0:
            self.logger.warning(str(errors))

        return errors




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Type

from neuro_san.internals.filters.maximal_message_filter import MaximalMessageFilter
from neuro_san.internals.filters.message_filter import MessageFilter
from neuro_san.internals.filters.minimal_message_filter import MinimalMessageFilter

TYPE_TO_MESSAGE_FILTER_CLASS: Dict[Any, Type[MessageFilter]] = {
    0:  MinimalMessageFilter,
    1:  MinimalMessageFilter,
    2:  MaximalMessageFilter,

    "UNKNOWN":  MinimalMessageFilter,
    "MINIMAL":  MinimalMessageFilter,
    "MAXIMAL":  MaximalMessageFilter,
}


class MessageFilterFactory:
    """
    Class for creating MessageFilters
    """

    @staticmethod
    def create_message_filter(chat_filter: Dict[str, Any]) -> MessageFilter:
        """
        :param chat_filter: The ChatFilter dictionary to process.
        :return: A MessageFilter that corresponds to the contents
        """

        # For now the default is MAXIMAL simply to emulate current behavior.
        # After the ChatFilter API gets released this will eventually change to "MINIMAL".
        default: str = "MINIMAL"
        chat_filter_type: Any = default

        # Get what was in the request
        if chat_filter is not None:
            chat_filter_type = chat_filter.get("chat_filter_type", chat_filter_type)

        # Change strings so they are suitable for lookup
        if isinstance(chat_filter_type, str):
            chat_filter_type = chat_filter_type.upper()

        chat_filter_class: MessageFilter = TYPE_TO_MESSAGE_FILTER_CLASS.get(chat_filter_type)
        if chat_filter_class is None:
            chat_filter_class = TYPE_TO_MESSAGE_FILTER_CLASS.get(default)

        # Instantiate the class and return
        message_filter: MessageFilter = chat_filter_class()
        return message_filter



# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from neuro_san import REGISTRIES_DIR
from neuro_san.interfaces.coded_tool import CodedTool
from neuro_san.interfaces.reservation import Reservation
from neuro_san.internals.graph.activations.branch_activation import BranchActivation
from neuro_san.internals.graph.persistence.agent_network_restorer import AgentNetworkRestorer
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.reservations.reservation_util import ReservationUtil


class Copyist(BranchActivation, CodedTool):
    """
    CodedTool implementation of a copyist for the copy_cat reservations example.

    The copyist needs at least a single argument, which is an existing "agent_name",
    reads that hocon file in and makes a temporary agent out of it.

    Optionally, there might be a "call_text" argument. If that is there, then immediately
    invoke the reservation's network with that input.

    The idea is merely to have a proof of concept surrounding the Reservations infrastructure.

    ---

    Upon activation by the agent hierarchy, a CodedTool will have its
    invoke() call called by the system.

    Implementations are expected to clean up after themselves.
    """

    # pylint: disable=too-many-locals
    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Any:
        """
        Called when the coded tool is invoked asynchronously by the agent hierarchy.
        Strongly consider overriding this method instead of the "easier" synchronous
        version above when the possibility of making any kind of call that could block
        (like sleep() or a socket read/write out to a web service) is within the
        scope of your CodedTool.

        :param args: An argument dictionary whose keys are the parameters
                to the coded tool and whose values are the values passed for them
                by the calling agent.  This dictionary is to be treated as read-only.
        :param sly_data: A dictionary whose keys are defined by the agent hierarchy,
                but whose values are meant to be kept out of the chat stream.

                This dictionary is largely to be treated as read-only.
                It is possible to add key/value pairs to this dict that do not
                yet exist as a bulletin board, as long as the responsibility
                for which coded_tool publishes new entries is well understood
                by the agent chain implementation and the coded_tool implementation
                adding the data is not invoke()-ed more than once.
        :return: A return value that goes into the chat stream.
        """
        copy_agent: str = args.get("agent_name")
        if copy_agent is None or len(copy_agent) == 0:
            return "Need a non-empty value for copy_agent"

        # Make sure we have a hocon file reference
        if not copy_agent.endswith(".hocon"):
            copy_agent = f"{copy_agent}.hocon"

        # Remove the .hocon suffix for this string
        use_agent_name: str = copy_agent[:-6]

        # Restore the given agent network to find its spec dictionary
        copy_file: str = REGISTRIES_DIR.get_file_in_basis(copy_agent)
        restorer = AgentNetworkRestorer()
        network: AgentNetwork = restorer.restore(file_reference=copy_file)
        my_agent_spec: Dict[str, Any] = network.get_config()

        # Creating Reservations can be done outside the with-statement
        lifetime_in_seconds: float = 5 * 60.0

        reservation: Reservation = None
        error: str = None
        reservation, error = await ReservationUtil.wait_for_one(
            args, my_agent_spec, lifetime_in_seconds, prefix=f"copy_cat-{use_agent_name}"
        )

        if error is not None:
            return error

        # Get info from the reservation
        reservation_id: str = reservation.get_reservation_id()
        lifetime_in_seconds: float = reservation.get_lifetime_in_seconds()

        # Put the output in sly_data for less LLM "telephone" interference
        sly_data["agent_reservations"] = [
            {
                "reservation_id": reservation_id,
                "lifetime_in_seconds": lifetime_in_seconds,
                "expiration_time_in_seconds": reservation.get_expiration_time_in_seconds()
            }
        ]

        if args.get("call_text") is None:
            # Don't call the agent we just created as a tool
            return (
                f"The temporary agent name is {reservation_id}."
                + f"Hurry, it's only available for {lifetime_in_seconds} seconds."
            )

        # Call the agent we just created for our reservation
        res_args: Dict[str, Any] = {"input": args.get("call_text")}
        res_return: str = await self.use_reservation(reservation_id, res_args, sly_data)
        return res_return




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.interfaces.async_hopper import AsyncHopper
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.base_message_dictionary_converter import BaseMessageDictionaryConverter


class MessageJournal(Journal):
    """
    Journal implementation for putting entries into a Hopper
    for storage for later processing.
    """

    def __init__(self, hopper: AsyncHopper):
        """
        Constructor

        :param hopper: A handle to an AsyncHopper implementation, onto which
                       any message will be put().
        """
        self.hopper: AsyncHopper = hopper

    async def write_message(self, message: BaseMessage, origin: List[Dict[str, Any]]):
        """
        Writes a BaseMessage entry into the journal
        :param message: The BaseMessage instance to write to the journal
        :param origin: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        converter = BaseMessageDictionaryConverter(origin=origin)
        message_dict: Dict[str, Any] = converter.to_dict(message)

        # Queue Producer from this:
        #   https://stackoverflow.com/questions/74130544/asyncio-yielding-results-from-multiple-futures-as-they-arrive
        # The synchronous=True is necessary when an async HTTP request is at the get()-ing end of the queue,
        # as the journal messages come from inside a separate event loop from that request. The lock
        # taken here ends up being harmless in the synchronous request case (like for gRPC) because
        # we would only be blocking our own event loop.
        await self.hopper.put(message_dict, synchronous=True)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any

from neuro_san import REGISTRIES_DIR
from neuro_san.client.simple_one_shot import SimpleOneShot
from neuro_san.test.evaluators.abstract_agent_evaluator import AbstractAgentEvaluator
from neuro_san.test.interfaces.assert_forwarder import AssertForwarder


class GistAgentEvaluator(AbstractAgentEvaluator):
    """
    AbstractAgentEvaluator implementation that looks asks an LLM if the test values
    match the gist of the value to be verified against.
    """

    def __init__(self, asserts: AssertForwarder, negate: bool = False,
                 discriminator_agent: str = None, connection_type: str = "direct"):
        """
        Constructor

        :param asserts: The AssertForwarder instance to handle failures
        :param negate: If true the verification criteria should be negated
        :param discriminator_agent: The name of the discriminator agent to use for the test
                                By default this is None, implying the stock "gist.hocon"
                                agent shipped with neuro-san should be used.
        :param connection_type: The string connection type to pass to the AgentSessionFactory
        """
        super().__init__(asserts, negate)
        self.connection_type: str = connection_type
        self.discriminator_agent: str = discriminator_agent

        if discriminator_agent is None:
            self.discriminator_agent = REGISTRIES_DIR.get_file_in_basis("gist.hocon")

    def test_one(self, verify_value: Any, test_value: Any):
        """
        :param verify_value: The value to verify against
        :param test_value: The value appearing in the test sample
        """

        acceptance_criteria: str = str(verify_value)
        text_sample: str = str(test_value)
        pass_fail: bool = self.ask_llm(acceptance_criteria=acceptance_criteria,
                                       text_sample=text_sample)

        if self.negate:
            self.asserts.assertNotGist(pass_fail, acceptance_criteria, text_sample)
        else:
            self.asserts.assertGist(pass_fail, acceptance_criteria, text_sample)

    def ask_llm(self, acceptance_criteria: str, text_sample: str) -> bool:
        """
        Asks an llm if a text_sample meets a description of acceptance_criteria

        :param acceptance_criteria: String description of the acceptance criteria
                    of the text sample
        :param text_sample: A sample of text to evaluate
        :return: A boolean value as to whether or not the text_sample passes the acceptance_criteria.
        """
        text: str = f"""
The acceptance_criteria is:
"{acceptance_criteria}".

The text_sample is:
"{text_sample}".
"""

        # Use the "gist" agent to do the evalution
        one_shot = SimpleOneShot(self.discriminator_agent, self.connection_type)
        raw_answer: str = one_shot.get_answer_for(text)
        answer: str = raw_answer.lower()
        test_passes: bool = self.determine_pass_fail(answer)
        return test_passes

    def determine_pass_fail(self, answer: str) -> bool:
        """
        :param answer: The answer from the conversation with the LLM
        :return: True if the test had passed.
        """
        # Determine pass/fail by examining the answer
        passing: bool = "pass" in answer
        failing: bool = "fail" in answer

        # Be sure there is no equivocating.
        test_passes: bool = passing and not failing
        test_fails: bool = failing and not passing
        only_one: bool = test_passes or test_fails

        # Specifically use assertEqual() here to reserve assertTrue/False for
        # whether or not the test itself passed, as those feed into the Assessor.
        self.asserts.assertEqual(only_one, True)

        return test_passes




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Dict
from typing import List

import logging
import threading

from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.interfaces.agent_state_listener import AgentStateListener
from neuro_san.internals.interfaces.agent_storage_source import AgentStorageSource
from neuro_san.internals.network_providers.single_agent_network_provider import SingleAgentNetworkProvider


class AgentNetworkStorage(AgentStorageSource):
    """
    Service-wide storage for AgentNetworkProviders containing
    a table of currently active AgentNetworks for each agent registered to the service.
    Note: a mapping from an agent to its AgentNetwork is dynamic,
          as it is possible to change agents definitions at service run-time.
    """

    def __init__(self):
        self.agents_table: Dict[str, AgentNetwork] = {}
        self.logger = logging.getLogger(self.__class__.__name__)
        self.lock = threading.Lock()
        self.listeners: List[AgentStateListener] = []

    def add_listener(self, listener: AgentStateListener):
        """
        Add a state listener to be notified when status of service agents changes.
        """
        self.listeners.append(listener)

    def remove_listener(self, listener: AgentStateListener):
        """
        Remove a state listener from registered set.
        """
        if listener in self.listeners:
            self.listeners.remove(listener)

    def add_agent_network(self, agent_name: str, agent_network: AgentNetwork):
        """
        This method is a single point of entry in the service
        where we register a new agent name -> AgentNetwork pair in the service scope
        or notify the service that for existing agent its AgentNetwork has been modified.
        """
        is_new: bool = False
        with self.lock:
            is_new = self.agents_table.get(agent_name) is None
            self.agents_table[agent_name] = agent_network

        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            if is_new:
                listener.agent_added(agent_name, self)
                self.logger.info("ADDED network for agent %s : %d", agent_name, id(agent_network))
            else:
                listener.agent_modified(agent_name, self)
                self.logger.info("REPLACED network for agent %s : %d", agent_name, id(agent_network))

    def setup_agent_networks(self, agent_networks: Dict[str, AgentNetwork]):
        """
        Replace agents networks with a new collection.
        Previous state could be empty.
        """
        current_agents = set(self.agents_table.keys())
        new_agents = set(agent_networks.keys())
        # Remove agents which are not in the new collection:
        agents_to_remove = current_agents - new_agents
        for agent_name in agents_to_remove:
            self.remove_agent_network(agent_name)
        # Now add (or possibly replace) agents from new collection:
        for agent_name in new_agents:
            self.add_agent_network(agent_name, agent_networks[agent_name])

    def remove_agent_network(self, agent_name: str):
        """
        Remove agent name and its AgentNetwork from service scope,
        so that agent becomes unavailable on our server.
        """
        with self.lock:
            agent_network: AgentNetwork = self.agents_table.get(agent_name, None)
            self.agents_table.pop(agent_name, None)

        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            listener.agent_removed(agent_name, self)
        # We can log id(None) if necessary - that's a legal value
        self.logger.info("REMOVED network for agent %s : %d", agent_name, id(agent_network))

    def get_agent_network_provider(self, agent_name: str) -> AgentNetworkProvider:
        """
        Get AgentNetworkProvider for a specific agent
        :param agent_name: name of an agent
        """
        return SingleAgentNetworkProvider(agent_name, self.agents_table)

    def get_agent_names(self) -> List[str]:
        """
        Return static list of agent names.
        """
        with self.lock:
            # Create static snapshot of agents names collection
            return list(self.agents_table.keys())




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
import json

from neuro_san import TOP_LEVEL_DIR
from neuro_san.internals.interfaces.dictionary_validator import DictionaryValidator
from neuro_san.service.mcp.validation.mcp_request_validator import McpRequestValidator
from neuro_san.service.mcp.interfaces.client_session_policy import ClientSessionPolicy
from neuro_san.service.mcp.session.mcp_no_sessions_policy import McpNoSessionsPolicy
from neuro_san.service.mcp.util.mcp_request_util import McpRequestUtil


class McpServerContext:
    """
    Class representing the server run-time context,
    necessary for handling MCP clients requests.
    """

    def __init__(self):
        self.protocol_schema_filepath = None
        self.protocol_schema = None
        self.session_policy = None
        self.request_validator = None
        self.enabled: bool = False

    def set_enabled(self, enabled: bool) -> None:
        """
        Enable or disable the MCP service.
        :param enabled: Flag indicating if the service should be enabled
        """
        if not self.enabled and enabled:
            print(">>>>>>>>>>>> Enabling MCP service...")
            # MCP service is being enabled, set it up:
            schema_name: str = f"service/mcp/validation/mcp-schema-{McpRequestUtil.get_mcp_version()}.json"
            self.protocol_schema_filepath = TOP_LEVEL_DIR.get_file_in_basis(schema_name)
            try:
                with open(self.protocol_schema_filepath, "r", encoding="utf-8") as schema_file:
                    self.protocol_schema = json.load(schema_file)
                    self.request_validator = McpRequestValidator(self.protocol_schema)
            except Exception as exc:  # pylint: disable=broad-exception-caught
                raise RuntimeError(f"Cannot load MCP protocol schema from "
                                   f"'{self.protocol_schema_filepath}': {str(exc)}") from exc
            # Create a new session manager:
            self.session_policy = McpNoSessionsPolicy()
        self.enabled = enabled

    def is_enabled(self) -> bool:
        """
        Check if the MCP service is enabled.
        :return: True if the service is enabled, False otherwise
        """
        return self.enabled

    def get_protocol_version(self) -> str:
        """
        Get the MCP protocol version supported by this service.
        :return: The MCP protocol version
        """
        return McpRequestUtil.get_mcp_version()

    def get_request_validator(self) -> DictionaryValidator:
        """
        Get the request validator for this context.
        :return: The request validator
        """
        return self.request_validator

    def get_session_policy(self) -> ClientSessionPolicy:
        """
        Get the MCP session policy for this context.
        :return: The session policy instance
        """
        return self.session_policy




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Generator

import json
import requests

from leaf_common.time.timeout import Timeout

from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.session.abstract_http_service_agent_session import AbstractHttpServiceAgentSession
from neuro_san.session.mcp_chat_response_dictionary_converter import McpChatResponseDictionaryConverter

# MCP protocol version supported by this MCP session
# Protocol specification is available at:
# https://modelcontextprotocol.io/specification/2025-06-18
MCP_VERSION: str = "2025-06-18"


class McpServiceAgentSession(AbstractHttpServiceAgentSession, AgentSession):
    """
    Implementation of AgentSession that talks to an MCP protocol service.
    This is largely only used by command-line tests.
    """
    MCP_PROTOCOL_VERSION: str = "MCP-Protocol-Version"

    # pylint: disable=too-many-arguments,too-many-positional-arguments, too-many-locals
    def __init__(self, host: str = None,
                 port: str = None,
                 timeout_in_seconds: int = 30,
                 metadata: Dict[str, str] = None,
                 security_cfg: Dict[str, Any] = None,
                 umbrella_timeout: Timeout = None,
                 streaming_timeout_in_seconds: int = None,
                 agent_name: str = None,
                 session_name: str = "Neuro SAN MCP Session"):
        """
        Creates an MCP protocol session.

        :param host: the service host to connect to
                        If None, will use a default
        :param port: the service port
                        If None, will use a default
        :param timeout_in_seconds: timeout to use when communicating
                        with the service
        :param metadata: A grpc metadata of key/value pairs to be inserted into
                         the header. Default is None. Preferred format is a
                         dictionary of string keys to string values.
        :param security_cfg: An optional dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  Default is None, uses insecure channel.
        :param umbrella_timeout: A Timeout object under which the length of all
                        looping and retries should be considered
        :param streaming_timeout_in_seconds: timeout to use when streaming to/from
                        the service. Default is None, indicating connection should
                        stay open until the (last) result is yielded.
        :param agent_name: The name of the agent to talk to
        :param session_name: The name of the session used in handshake exchange.
        """
        super().__init__(host=host, port=port, timeout_in_seconds=timeout_in_seconds,
                         metadata=metadata, security_cfg=security_cfg, umbrella_timeout=umbrella_timeout,
                         streaming_timeout_in_seconds=streaming_timeout_in_seconds, agent_name=agent_name)
        # Do initial handshake and protocol negotiation
        handshake_dict: Dict[str, Any] = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "initialize",
            "params": {
                "protocolVersion": MCP_VERSION,
                "capabilities": {
                    "roots": {
                        "listChanged": False
                    },
                    "sampling": {},
                    "elicitation": {}
                },
                "clientInfo": {
                    "name": session_name,
                    "title": session_name,
                    "version": "1.0.0"
                }
            }
        }
        headers: Dict[str, str] = self.get_headers()
        headers["Content-Type"] = "application/json"

        path: str = self.get_request_path("initialize")
        try:
            response = requests.post(path, json=handshake_dict, headers=headers, timeout=self.timeout_in_seconds)
            response.raise_for_status()
            response_dict = json.loads(response.text)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc

        # Extract the protocol version from the handshake response
        self.protocol_version: str = response_dict.get("result", {}).get("protocolVersion", None)

        # Confirm the protocol version is supported by this client
        if self.protocol_version not in [MCP_VERSION]:
            raise ValueError(f"Unsupported MCP protocol version: {self.protocol_version}")

        # Acknowledge the successful handshake
        ack_dict: Dict[str, Any] = {
            "jsonrpc": "2.0",
            "method": "notifications/initialized"
        }
        try:
            response = requests.post(path, json=ack_dict, headers=headers, timeout=self.timeout_in_seconds)
            response.raise_for_status()
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc

    def function(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the FunctionRequest
                    protobufs structure. Has the following keys:
                        <None>
        :return: A dictionary version of the FunctionResponse
                    protobufs structure. Has the following keys:
                "function" - the dictionary description of the function
        """
        # Get the list of tools available from the service
        request_dict: Dict[str, Any] = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "tools/list",
            "params": {
            }
        }
        headers: Dict[str, str] = self.get_headers()
        headers["Content-Type"] = "application/json"
        headers[self.MCP_PROTOCOL_VERSION] = self.protocol_version

        path: str = self.get_request_path("tools/list")
        try:
            response = requests.post(path, json=request_dict, headers=headers, timeout=self.timeout_in_seconds)
            response.raise_for_status()
            response_dict = json.loads(response.text)
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc

        tools_list: List[Dict[str, Any]] = response_dict.get("result", {}).get("tools", [])
        for tool in tools_list:
            name: str = tool.get("name", None)
            if name == self.agent_name:
                tool_description: str = tool.get("description", None)
                if tool_description is not None:
                    return {
                        "function": {"description": tool_description}
                    }

        return None

    def connectivity(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConnectivityRequest.
        :return: A dictionary version of the ConnectivityResponse.
        """
        # Not used in MCP protocol; return empty connectivity info
        request_dict: Dict[str, Any] = {}
        return request_dict

    def streaming_chat(self, request_dict: Dict[str, Any]) -> Generator[Dict[str, Any], None, None]:
        """
        :param request_dict: A dictionary version of the ChatRequest
                    protobufs structure. Has the following keys:
            "user_message" - A ChatMessage dict representing the user input to the chat stream
            "chat_context" - A ChatContext dict representing the state of the previous conversation
                            (if any)
        :return: An iterator of dictionary versions of the ChatResponse
                    protobufs structure. Has the following keys:
            "response" - An optional ChatMessage dictionary.  See chat.proto for details.

            Note that responses to the chat input might be numerous and will come as they
            are produced until the system decides there are no more messages to be sent.
        """
        # Pack the chat request dictionary into an MCP method call format:
        mcp_payload = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "tools/call",
            "params": {
                "name": self.agent_name,
                "arguments": request_dict,
            },
        }

        headers: Dict[str, str] = self.get_headers()
        headers["Content-Type"] = "application/json"
        headers[self.MCP_PROTOCOL_VERSION] = self.protocol_version

        path: str = self.get_request_path("streaming_chat")
        try:
            with requests.post(path, json=mcp_payload, headers=headers,
                               timeout=self.streaming_timeout_in_seconds) as response:
                response.raise_for_status()
                for line in response.iter_lines(decode_unicode=True):
                    if line.strip():  # Skip empty lines
                        # Each line is a JSON object representing an MCP tool call(chat) response
                        result_dict = json.loads(line)
                        result_dict = McpChatResponseDictionaryConverter().to_dict(result_dict)
                        yield result_dict
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc

    def get_request_path(self, method: str) -> str:
        """
        :param method: The method endpoint we wish to reach
        :return: The full URL for accessing the method, given the host, port and agent_name.
        """
        _ = method
        scheme: str = "http"
        if self.security_cfg is not None:
            scheme = "https"
        return f"{scheme}://{self.use_host}:{self.use_port}/mcp"




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from neuro_san.internals.graph.activations.abstract_class_activation import AbstractClassActivation


class ClassActivation(AbstractClassActivation):
    """
    A ClassActivation that retrieves the full class reference directly from the class specification
    in agent network hocon.
    """

    def get_full_class_ref(self) -> str:
        """
        Returns the full class reference path directly from the class specification.

        This implementation expects the fully qualified class name to be provided
        in the "class" field of the `agent_tool_spec` dictionary.

        :return: A dot-separated string representing the full class path.
        """
        return self.agent_tool_spec.get("class")




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict


class AgentProgressReporter:
    """
    An interface for CodedTools to be able to report on an agent network's progress.

    Instances get handed down via the CodedTool's arguments dictionary via the
    "progress_reporter" key via the invoke() or async_invoke() methods.

    Typically, progress is reported as a dictionary that is JSON-serializable
    and interpreted by the client on a per-agent-network basis in the structure.
    The simplest and most easily interpreted structure is simply to report a key of "progress"
    with a value of a float between 0.0 and 1.0, but other keys/values can be used to report
    (say) partial progress on a structure that is being built by the agent network so that it
    can be visualized by an in-the-know client.

    Text messages can also be sent as content, but more as differential comments.
    These are not recommended, as any given client may not be able to parse them very easily.
    """

    async def async_report_progress(self, structure: Dict[str, Any], content: str = ""):
        """
        Reports the structure and optional message to the chat message stream returned to the client
        To be used from within CodedTool.async_invoke().

        :param structure: The Dictionary instance to write as progress.
                        All keys and values must be JSON-serializable.
        :param content: An optional message to send to the client
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Type
from typing import Union

import traceback

from pydantic import ConfigDict

from langchain_classic.callbacks.tracers.logging import LoggingCallbackHandler
from langchain_core.agents import AgentFinish
from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.language_models.base import BaseLanguageModel
from langchain_core.messages.ai import AIMessage
from langchain_core.messages.base import BaseMessage
from langchain_core.runnables.base import Runnable
from langchain_core.runnables.utils import Input
from langchain_core.runnables.utils import Output

from leaf_common.config.resolver_util import ResolverUtil

from neuro_san.internals.errors.error_detector import ErrorDetector
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.origination import Origination
from neuro_san.internals.run_context.interfaces.tool_caller import ToolCaller
from neuro_san.internals.run_context.langchain.journaling.journaling_callback_handler import JournalingCallbackHandler
from neuro_san.internals.run_context.langchain.token_counting.langchain_token_counter import LangChainTokenCounter
from neuro_san.internals.run_context.langchain.tracing.neuro_san_runnable import NeuroSanRunnable
from neuro_san.internals.run_context.langchain.util.api_key_error_check import ApiKeyErrorCheck

MINUTES: float = 60.0

# Lazily import specific errors from llm providers
API_ERROR_TYPES: Tuple[Type[Any], ...] = ResolverUtil.create_type_tuple([
                                            "openai.APIError",
                                            "anthropic.APIError",
                                            "langchain_google_genai.chat_models.ChatGoogleGenerativeAIError",
                                         ])


class RunContextRunnable(NeuroSanRunnable):
    """
    RunnablePassthrough implementation that intercepts journal messages
    """

    # Declarations of member variables here satisfy Pydantic style,
    # which is a type validator that langchain is based on which
    # is able to use JSON schema definitions to validate fields.
    agent_chain: Runnable

    primary_llm: BaseLanguageModel

    journal: Journal

    tool_caller: ToolCaller

    error_detector: ErrorDetector

    session_id: str

    # This guy needs to be a pydantic class and in order to have
    # a non-pydantic Journal as a member, we need to do this.
    model_config = ConfigDict(arbitrary_types_allowed=True)

    # pylint: disable=redefined-builtin
    async def run_it(self, inputs: Input) -> Output:
        """
        Transform a single input into an output.

        Args:
            inputs: The input to the `Runnable`.

        Returns:
            The output of the `Runnable`.
        """

        # Create an agent executor and invoke it with the most recent human message
        # as input.
        agent_spec: Dict[str, Any] = self.tool_caller.get_agent_tool_spec()

        max_execution_seconds: float = agent_spec.get("max_execution_seconds", 2.0 * MINUTES)

        # Per advice from https://python.langchain.com/docs/how_to/migrate_agent/#max_iterations
        max_iterations: int = agent_spec.get("max_iterations", 20)
        recursion_limit: int = max_iterations * 2 + 1

        # Create the list of callbacks to pass when invoking
        parent_origin: List[Dict[str, Any]] = self.origin
        base_journal: Journal = self.invocation_context.get_journal()
        origination: Origination = self.invocation_context.get_origination()
        callbacks: List[BaseCallbackHandler] = [
            JournalingCallbackHandler(self.journal, base_journal, parent_origin, origination)
        ]

        # Consult the agent spec for level of verbosity as it pertains to callbacks.
        agent_spec: Dict[str, Any] = self.tool_caller.get_agent_tool_spec()
        verbose: Union[bool, str] = agent_spec.get("verbose", False)
        if isinstance(verbose, str) and verbose.lower() in ("true", "extra", "logging"):
            # This particular class adds a *lot* of very detailed messages
            # to the logs.  Add this because some people are interested in it.
            callbacks.append(LoggingCallbackHandler(self.logger))

        runnable_config: Dict[str, Any] = self.prepare_runnable_config(callbacks=callbacks,
                                                                       recursion_limit=recursion_limit)

        # Attempt to count tokens/costs while invoking the agent.
        token_counter = LangChainTokenCounter(self.primary_llm, self.invocation_context, self.journal, self.origin)
        await token_counter.count_tokens(self.invoke_agent_chain(inputs, runnable_config), max_execution_seconds)

        return inputs

    # pylint: disable=too-many-locals
    async def invoke_agent_chain(self, inputs: Dict[str, Any], runnable_config: Dict[str, Any]):
        """
        Set the agent in motion

        :param inputs: The inputs to the agent_executor
        :param runnable_config: The runnable_config to send to the agent_executor
        """
        chain_result: Union[Dict[str, Any], AgentFinish, AIMessage] = None
        retries: int = 3
        exception: Exception = None
        backtrace: str = None
        while chain_result is None and retries > 0:
            try:
                chain_result: Dict[str, Any] = await self.agent_chain.ainvoke(input=inputs, config=runnable_config)
            except API_ERROR_TYPES as api_error:
                backtrace = traceback.format_exc()
                message: str = None
                if not ApiKeyErrorCheck.check_for_internal_error(backtrace):
                    # Does not look like internal LLM stack error:
                    message = ApiKeyErrorCheck.check_for_api_key_exception(api_error)
                if message is not None:
                    # Construct a uniform message to return to the client
                    # indicating that they likely have an API key problem,
                    # rather than retrying and hitting the same error again.
                    exception = None
                    backtrace = None
                    chain_result = {
                        "output": message
                    }
                    # Log the error with technical details for debugging purposes,
                    # but we are returning a more user-friendly message to the client.
                    self.logger.error("API KEY error detected: %s", str(api_error))
                    break
                # Continue with regular retry logic:
                self.logger.warning("retrying from %s", api_error.__class__.__name__)
                retries = retries - 1
                exception = api_error
            except KeyError as key_error:
                self.logger.warning("retrying from KeyError")
                retries = retries - 1
                exception = key_error
                backtrace = traceback.format_exc()
            except ValueError as value_error:
                response = str(value_error)
                find_string = "An output parsing error occurred. " + \
                              "In order to pass this error back to the agent and have it try again, " + \
                              "pass `handle_parsing_errors=True` to the AgentExecutor. " + \
                              "This is the error: Could not parse LLM output: `"
                if response.startswith(find_string):
                    # Agent is returning good stuff, but langchain is erroring out over it.
                    # From: https://github.com/langchain-ai/langchain/issues/1358#issuecomment-1486132587
                    # Per thread consensus, this is hacky and there are better ways to go,
                    # but removes immediate impediments.
                    chain_result = {
                        "output": response.removeprefix(find_string).removesuffix("`")
                    }
                else:
                    self.logger.warning("retrying from ValueError")
                    retries = retries - 1
                    exception = value_error
                    backtrace = traceback.format_exc()
            # pylint: disable=broad-exception-caught
            except Exception as exception_error:
                # This catches any errors from running middlewares and also error form exceeding the recursion_limit.
                self.logger.error("Got exception in  %s. Error: %s",
                                  self.__class__.__name__,
                                  exception_error,
                                  )
                # These are likely real issues and non-retryable.
                retries = 0
                exception = exception_error
                backtrace = traceback.format_exc()

        output: str = self.parse_chain_result(chain_result, exception, backtrace)
        return_message: BaseMessage = AIMessage(output)

        # Chat history is updated in write_message
        await self.journal.write_message(return_message)

    def parse_chain_result(self, chain_result: Union[Dict[str, Any], AgentFinish, AIMessage],
                           exception: Exception, backtrace: str) -> str:
        """
        Parse the result from the langchain chain.

        :param chain_result: The result from invoking the agent chain.
                        Can be:
                        * An AgentFinish instance whose return_values can be any one of the following
                        * A dictionary whose keys might be:
                            "output" - the actual output to use
                            "messages" - effectively a chat history
                        * An AIMessage whose content is the output to use
        :param exception: Any exception that happened along the way
        :param backtrace: Any backtrace to the exception that happened along the way
        :return: A string value to return as the result of the run.
        """

        # Initialize our output.
        # The value here might morph a bit between types, but when we return
        # something we expect it to be a string.
        output: Union[str, List[Dict[str, Any]]] = None

        if chain_result is None and exception is not None:
            # We got an exception instead of a proper result. Say so.
            output = f"Agent stopped due to exception {exception}"
        else:
            # Set some stuff up for later
            backtrace = None
            ai_message: AIMessage = None

            # Handle the AgentFinish case.
            # The return_values from there contain our output whether in string or dict form.
            # ??? From what path does this come?
            if isinstance(chain_result, AgentFinish):
                chain_result = chain_result.return_values

            if isinstance(chain_result, Dict):
                # Normal return value from a chain is a dict.
                # The dict in question usually has chat history in a messages field.
                # We want the last AIMessage from that chat history.
                messages: List[BaseMessage] = chain_result.get("messages", [])
                for message in reversed(messages):
                    if isinstance(message, AIMessage):
                        ai_message = message
                        break

                if ai_message is None:
                    # We didn't find an AIMessage, so look for straight-up output key
                    output = chain_result.get("output")

            elif isinstance(chain_result, AIMessage):
                # Sometimes we get an AIMessage from a tool call.
                ai_message = chain_result

            if ai_message is not None:
                # We generally want the content of any single AIMessage we found from above
                output = ai_message.content

        # In general, output is a string. but output from Anthropic can either be
        # a single string or a list of content blocks.
        # If it is a list, "text" is a key of a dictionary which is the first element of
        # the list. For more details: https://python.langchain.com/docs/integrations/chat/anthropic/#content-blocks
        if isinstance(output, list):
            output = output[0].get("text", "")

        # See if we had some kind of error and format accordingly, if asked for.
        output = self.error_detector.handle_error(output, backtrace)
        return output




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from __future__ import annotations

from typing import Any
from typing import Dict
from typing import Tuple

from neuro_san.interfaces.reservation import Reservation


class ReservationsStorage:
    """
    An interface for implementations of basic Reservations storage,
    supporting addition Reservations in bulk and retrieval of individual Reservations,
    as well as expiration of Reservations based on their lifetime.
    """

    def add_reservations(self, reservations_dict: Dict[Reservation, Any],
                         source: str = None):
        """
        Add a set of reservations for agent networks en-masse

        :param reservations_dict: A mapping of Reservation -> some deployable entity
        :param source: A string describing where the deployment was coming from
        """
        raise NotImplementedError

    def get_one_reservation(self, obj_key: str) -> Tuple[Reservation, Any]:
        """
        Extract a single reservation.

        :param obj_key: unique key for the reservation
        :return: Tuple of (reservation, agent data) if successful
                 and reservation is not expired,
                 (None, None) otherwise
        """
        raise NotImplementedError

    def expire_reservations(self):
        """
        Remove Reservations that are expired
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict

from neuro_san.service.http.logging.http_logger import HttpLogger
from neuro_san.service.mcp.util.mcp_request_util import McpRequestUtil


class McpPromptsProcessor:
    """
    Class implementing "prompts"-related MCP requests.
    https://modelcontextprotocol.io/specification/2025-06-18/server/prompts
    """

    def __init__(self, logger: HttpLogger):
        self.logger: HttpLogger = logger

    async def list_prompts(self, request_id, metadata: Dict[str, Any]) -> Dict[str, Any]:
        """
        List available MCP promp templates.
        :param request_id: MCP request id;
        :param metadata: http-level request metadata;
        :return: json dictionary with prompts list in MCP format
        """
        _ = metadata
        return {
            "jsonrpc": "2.0",
            "id": McpRequestUtil.safe_request_id(request_id),
            "result": {
                "prompts": []
            }
        }




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any

from neuro_san.interfaces.async_agent_session import AsyncAgentSession


class AsyncAgentSessionFactory:
    """
    Creates asynchronous AsyncAgentSessions for external agents.
    """

    def create_session(self, agent_url: str, invocation_context: Any, invocation: str = None) -> AsyncAgentSession:
        """
        :param agent_url: A url string pointing to an external agent that came from
                    a tools list in an agent spec.
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.

                    Note: At this interface level we are typing this as Any to avoid
                    an import cycle.  This will always be an InvocationContext.

        :param invocation: String describing how the agent wants to be invoked.
                            Can be: "chatbot" - implies waiting for an answer.
                                    "event" - implies no answer needed
                                    None - implies chatbot

        :return: An implementation of AsyncAgentSession through which
                 communications about external agents can be made.
        """
        raise NotImplementedError

    def is_use_direct(self) -> bool:
        """
        :return: When True, will use a Direct session for external agents that would reside on the same server.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from uuid import uuid4
from uuid import UUID

from os import environ

from neuro_san.interfaces.reservation import Reservation


class AgentReservation(Reservation):
    """
    A Reservation for an Agent (network).
    """

    def __init__(self, lifetime_in_seconds: float, prefix: str = ""):
        """
        Constructor

        :param lifetime_in_seconds: The number of seconds the reservation is allowed to exist.
        :param prefix: A string prefix to prepend to the id so as to provide external context.
        """
        super().__init__(lifetime_in_seconds)

        # Use a uuid. Add a prefix if one is provided
        self.id: str = str(uuid4())
        self.prefix: str = prefix
        if self.prefix is None:
            self.prefix = ""
        if len(self.prefix) > 0:
            self.prefix = f"{self.prefix}-"

    def get_reservation_id(self) -> str:
        """
        :return: The id associated with the reservation instance
        """
        return f"{self.prefix}{self.id}"

    @staticmethod
    def is_reservation_name(agent_name: str) -> bool:
        """
        :param agent_name: The name of an agent
        :return: True if the name is in the format of a reservation id, False otherwise.
        """
        # We expect the format to be <prefix>-<uuid>, where the prefix can be empty.
        if not isinstance(agent_name, str):
            return False
        # Get uuid part without possible prefix:
        uuid_part: str = agent_name[-36:]
        try:
            u = UUID(uuid_part)
            if u.version == 4:
                return True
            return False
        except ValueError:
            return False

    def get_prefix(self) -> str:
        """
        :return: The prefix assigned at construct time
        """
        return self.prefix

    def get_url(self) -> str:
        """
        :return: A url associated with the reservation.
                 Can be None if this is not an option for what we are reserving.
        """
        # Start with a locally bound name.
        url: str = self.get_reservation_id()

        # See if we have an external agent name to tack on.
        external_url: str = environ.get("AGENT_EXTERNAL_SERVER_URL")
        if external_url is not None:

            # Remove any ending slashes to standardize
            while external_url.endswith("/"):
                external_url = external_url[:-1]

            # We still have something, so prepend on the external_url
            if len(external_url) > 0:
                # Assume always http for now.
                url = f"{external_url}/api/v1/{url}"

        return url

    def set_expiration_from(self, use_now_in_seconds: float,
                            max_lifetime_in_seconds: float):
        """
        Set the expiration time in seconds using the input as a basis for "now"..
        """
        actual_lifetime: float = min(self.get_lifetime_in_seconds(), max_lifetime_in_seconds)
        self.expiration_time_in_seconds = use_now_in_seconds + actual_lifetime




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

import asyncio
import logging
from time import time
from typing import Any
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing_extensions import override

from langchain_community.callbacks.bedrock_anthropic_callback import MODEL_COST_PER_1K_INPUT_TOKENS
from langchain_community.callbacks.bedrock_anthropic_callback import MODEL_COST_PER_1K_OUTPUT_TOKENS
from langchain_community.callbacks.openai_info import get_openai_token_cost_for_model
from langchain_community.callbacks.openai_info import TokenType
from langchain_core.callbacks import AsyncCallbackHandler
from langchain_core.messages import AIMessage
from langchain_core.messages import BaseMessage
from langchain_core.messages.ai import UsageMetadata
from langchain_core.outputs import ChatGeneration, LLMResult

EMPTY = ""
CLASS_TABLE = {
    # Chat model class : Provider class
    "AzureChatOpenAI": "azure-openai",
    "ChatAnthropic": "anthropic",
    "ChatBedrock": "bedrock",
    "ChatGoogleGenerativeAI": "gemini",
    "ChatNVIDIA": "nvidia",
    "ChatOllama": "ollama",
    "ChatOpenAI": "openai",
}


# pylint: disable=too-many-ancestors
# pylint: disable=too-many-instance-attributes
class LlmTokenCallbackHandler(AsyncCallbackHandler):
    """
    Callback handler that tracks token usage via "AIMessage.usage_metadata".

    This class is a modification of LangChain’s "UsageMetadataCallbackHandler" and "OpenAICallbackHandler":
    - https://python.langchain.com/api_reference/_modules/langchain_core/callbacks/usage.html
    #get_usage_metadata_callback
    - https://python.langchain.com/api_reference/_modules/langchain_community/callbacks/openai_info.html
    #OpenAICallbackHandler

    It collects token usage from the "usage_metadata" field of "AIMessage" each time an LLM or chat model
    finishes execution.
    The metadata is a dictionary that may include:
    - "input_tokens" (collected as "prompt_tokens")
    - "output_tokens" (collected as "completion_tokens")
    - "total_tokens"

    This handler tracks these values internally and is compatible with models that populate "usage_metadata",
    regardless of provider.

    Note:
    Token cost is calculated using prices from the LLM info file when available.
    For OpenAI, Azure OpenAI, Anthropic, and Bedrock Anthropic, there are  fallbacks to lookup tables in:
        - langchain_community.callbacks.openai_info.py
        - langchain_community.callbacks.bedrock_anthropic_callback.py
    If no price information is found, the cost defaults to 0.
    """

    # Token stats
    total_tokens: int = 0
    prompt_tokens: int = 0
    completion_tokens: int = 0
    successful_requests: int = 0
    total_cost: float = 0.0

    def __init__(self, llm_infos: Dict[str, Any]):
        """Initialize the CallbackHandler."""
        super().__init__()
        self._lock = asyncio.Lock()
        self.llm_infos: Dict[str, Any] = llm_infos
        self.provider_class: str = None
        self.start_time: float = None

        # Dictionary for accumulating token stats of models. For example
        # {"openai": {"gpt-4o": {"total_tokens": 100, "prompt_tokens": 80, ...}, "gpt_4.1": {...}}, }
        # Note that models with the same name but different providers counts as different models.
        self.models_token_dict: Dict[str, Any] = {}

    @override
    def __repr__(self) -> str:
        return (
            f"Tokens Used: {self.total_tokens}\n"
            f"\tPrompt Tokens: {self.prompt_tokens}\n"
            f"\tCompletion Tokens: {self.completion_tokens}\n"
            f"Successful Requests: {self.successful_requests}\n"
            f"Total Cost (USD): ${self.total_cost}\n"
            f"Model Info: {self.models_token_dict}"
        )

    @override
    async def on_chat_model_start(
        self,
        serialized: dict[str, Any],
        messages: list[list[BaseMessage]],
        **kwargs: Any
    ):
        """
        Extract the LLM class and start timer when chat model starts.
        :param serialized: Dictionary of metadata of the invoked model
        """
        # Chat moddel class of the LLM is in the last item of the id list
        chat_model_class: str = serialized.get("id")[-1]
        # Match the chat model class with neuro-san model class
        self.provider_class = CLASS_TABLE.get(chat_model_class)
        # If no match found, use chat model class instead
        if not self.provider_class:
            self.provider_class = chat_model_class
        if self.provider_class not in self.models_token_dict:
            self.models_token_dict[self.provider_class] = {}

        # Start timer
        self.start_time = time()

    @override
    async def on_llm_end(self, response: LLMResult, **kwargs: Any):
        """
        Collect token usage when llm ends.
        :param response: Output from chat model
        """
        # Calculate time latency for each llm
        # Note that this will be slightly lower time taken by the agent
        time_taken_in_seconds: float = time() - self.start_time

        # Check for usage_metadata (Only work for langchain-core >= 0.2.2)
        try:
            generation = response.generations[0][0]
        except IndexError:
            generation = None

        usage_metadata: UsageMetadata = None
        response_metadata: Dict[str, Any] = None
        model_name: str = EMPTY
        if isinstance(generation, ChatGeneration):
            try:
                message = generation.message
                if isinstance(message, AIMessage):
                    # Token info is in an attribute of AIMessage called "usage_metadata".
                    usage_metadata = message.usage_metadata
                    # Get model name so that cost can be determined if needed.
                    response_metadata = message.response_metadata
                    if response_metadata:
                        if "model_name" in response_metadata:
                            model_name = response_metadata.get("model_name")
                        elif "model_id" in response_metadata:
                            model_name = response_metadata.get("model_id")
                        elif "model" in response_metadata:
                            model_name = response_metadata.get("model")
            except AttributeError:
                pass

        if usage_metadata:
            total_tokens: int = usage_metadata.get("total_tokens", 0)
            completion_tokens: int = usage_metadata.get("output_tokens", 0)
            prompt_tokens: int = usage_metadata.get("input_tokens", 0)

            # Calculate the total cost
            total_cost: float = self.calculate_token_costs(model_name, completion_tokens, prompt_tokens)

            # Update shared state behind lock
            async with self._lock:
                # Initialize model entry if this is the first time we see this model
                if model_name not in self.models_token_dict[self.provider_class]:
                    self._init_model_entry(model_name)

                # Update per-model stats.
                self.models_token_dict[self.provider_class][model_name]["total_tokens"] += total_tokens
                self.models_token_dict[self.provider_class][model_name]["prompt_tokens"] += prompt_tokens
                self.models_token_dict[self.provider_class][model_name]["completion_tokens"] += \
                    completion_tokens
                self.models_token_dict[self.provider_class][model_name]["successful_requests"] += 1
                self.models_token_dict[self.provider_class][model_name]["total_cost"] += total_cost
                self.models_token_dict[self.provider_class][model_name]["time_taken_in_seconds"] += \
                    time_taken_in_seconds

                # Update per-agent stats
                self.total_tokens += total_tokens
                self.prompt_tokens += prompt_tokens
                self.completion_tokens += completion_tokens
                self.successful_requests += 1
                self.total_cost += total_cost

    def calculate_token_costs(self, model_name: str, completion_tokens: int, prompt_tokens: int) -> float:
        """
        Calculate token costs with fallback methods for different providers.
        :param model_name: Model to calculate the cost
        :param completion_tokens: Number of output tokens
        :param prompt_tokens: Number of input tokens
        :return: Total cost
        """

        # Try to get costs from llm_infos first
        completion_token_cost: float = \
            self._get_cost_from_info(model_name, completion_tokens, "price_per_1k_output_tokens")
        prompt_token_cost: float = self._get_cost_from_info(model_name, prompt_tokens, "price_per_1k_input_tokens")

        # Fallback to provider-specific methods for anthropic and openai based models
        # Since there are lookup tables for these models in langchain-community
        if self.provider_class in ["azure-openai", "openai"]:
            completion_token_cost = completion_token_cost or \
                self._get_openai_cost(model_name, completion_tokens, token_type=TokenType.COMPLETION)
            prompt_token_cost = prompt_token_cost or \
                self._get_openai_cost(model_name, prompt_tokens, token_type=TokenType.PROMPT)

        elif self.provider_class in ["anthropic", "bedrock"]:
            completion_token_cost = completion_token_cost or \
                self._get_anthropic_cost(model_name, completion_tokens, "completion")
            prompt_token_cost = prompt_token_cost or self._get_anthropic_cost(model_name, prompt_tokens, "prompt")

        # Return total cost
        return (completion_token_cost or 0.0) + (prompt_token_cost or 0.0)

    def _get_cost_from_info(self, model_name: str, num_tokens: int, price_key: str) -> Optional[float]:
        """
        Get cost from llm_infos if available.
        :param model_name: Model to calculate the cost
        :param num_tokens: Amount of tokens
        :param price_key: keyword to look in llm info for price
        :return: Token cost
        """
        if model_name not in self.llm_infos:
            return None

        price = self.llm_infos.get(model_name).get(price_key)
        return (num_tokens / 1000) * price if price is not None else None

    def _get_openai_cost(self, model_name: str, num_tokens: int, token_type: TokenType) -> Optional[float]:
        """
        Get OpenAI cost with error handling.
        :param model_name: Anthorpic model to calculate the cost
        :param num_tokens: Amount of tokens
        :param token_type: Type of token, either "prompt" (input) or "completion" (output)
        :return: Token cost
        """
        try:
            return get_openai_token_cost_for_model(model_name=model_name, num_tokens=num_tokens, token_type=token_type)
        except ValueError:
            return None

    def _get_anthropic_cost(self, model_name: str, num_tokens: int, token_type: str) -> Optional[float]:
        """
        Get Anthropic/Bedrock cost with error handling.
        :param model_name: Anthorpic model to calculate the cost
        :param num_tokens: Amount of tokens
        :param token_type: Type of token, either "prompt" (input) or "completion" (output)
        :return: Token cost
        """
        try:
            return self._get_anthropic_bedrock_token_cost(model_name, num_tokens, token_type)
        except ValueError:
            return None

    def _init_model_entry(self, model_name: str):
        """
        Initialize a new model entry in the tracking dictionary.
        :param model_name: LLM model name to put in the dictionary
        """
        self.models_token_dict[self.provider_class][model_name] = {
            "total_tokens": 0,
            "prompt_tokens": 0,
            "completion_tokens": 0,
            "successful_requests": 0,
            "total_cost": 0.0,
            "time_taken_in_seconds": 0.0
        }

    def _get_anthropic_bedrock_token_cost(
            self,
            model_name: str,
            num_tokens: int,
            token_type: Literal["completion", "prompt"] = "prompt"
    ) -> float:
        """
        Calculate token cost for Anthropic/Bedrock models from the lookup table with unified logic.
        :param model_name: Anthorpic model to calculate the cost
        :param num_tokens: Amount of tokens
        :param token_type: Type of token, either "prompt" (input) or "completion" (output)
        :return: Token cost
        """

        # Normalize model name for lookup
        normalized_model: str = self._normalize_model_name(model_name)

        # Find matching model in cost tables
        # matching_models = [model for model in MODEL_COST_PER_1K_INPUT_TOKENS if normalized_model in model]
        matching_models: List[str] = []
        for model in MODEL_COST_PER_1K_INPUT_TOKENS:
            if normalized_model in model:
                matching_models.append(model)

        if not matching_models:
            error_msg = f"Unknown model: {model_name}. Known models: {', '.join(MODEL_COST_PER_1K_INPUT_TOKENS.keys())}"
            logging.warning(error_msg)
            raise ValueError(error_msg)

        if len(matching_models) > 1:
            error_msg = f"Ambiguous model name '{model_name}'. Matches: {', '.join(matching_models)}"
            logging.warning(error_msg)
            raise ValueError(error_msg)

        # Calculate cost
        full_model_id: str = matching_models[0]
        if token_type == "prompt":
            cost_table: Dict[str, float] = MODEL_COST_PER_1K_INPUT_TOKENS
        else:
            cost_table = MODEL_COST_PER_1K_OUTPUT_TOKENS

        return (num_tokens / 1000) * cost_table[full_model_id]

    def _normalize_model_name(self, model_name: str) -> str:
        """
        Normalize model name for consistent lookup across Bedrock and Anthropic formats.
        :param model_name: Full name or id of Anthropic LLM
        :return: Name for checking in the lookup table
        """
        if "anthropic" in model_name:
            # For Bedrock: extract base model from cross-region inference profile
            # e.g., 'us.anthropic.claude-3-sonnet' -> 'anthropic.claude-3-sonnet'
            parts = model_name.split(".")
            if len(parts) >= 2:
                return ".".join(parts[-2:])

        return model_name




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.base import BaseMessage


class Journal:
    """
    An interface for journaling chat messages.
    """

    async def write_message(self, message: BaseMessage, origin: List[Dict[str, Any]]):
        """
        Writes a BaseMessage entry into the journal
        :param message: The BaseMessage instance to write to the journal
        :param origin: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Awaitable
from typing import Dict
from typing import List
from typing import Union

from asyncio import Task
from asyncio import TimeoutError as AsyncTimeout
from asyncio import wait_for
from contextvars import Context
from contextvars import ContextVar
from contextvars import copy_context
from time import time

from langchain_core.callbacks import AsyncCallbackHandler
from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.language_models.base import BaseLanguageModel

from leaf_common.asyncio.asyncio_executor import AsyncioExecutor
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.agent_message import AgentMessage
from neuro_san.internals.messages.origination import Origination
from neuro_san.internals.run_context.langchain.token_counting.get_llm_token_callback import get_llm_token_callback
from neuro_san.internals.run_context.langchain.token_counting.get_llm_token_callback import llm_token_callback_var

# Keep a ContextVar for the origin info.  We do this because the
# langchain callbacks this stuff is based on also uses ContextVars
# and we want to be sure these are in sync.
# See: https://docs.python.org/3/library/contextvars.html
ORIGIN_INFO: ContextVar[str] = ContextVar('origin_info', default=None)


class LangChainTokenCounter:
    """
    Helps with per-llm means of counting tokens.
    Main entrypoint is count_tokens().

    Notes as to how each BaseLanguageModel/BaseChatModel should be configured
    are in get_callback_for_llm()
    """

    def __init__(self, llm: BaseLanguageModel,
                 invocation_context: InvocationContext,
                 journal: Journal,
                 origin: List[Dict[str, Any]]):
        """
        Constructor

        :param llm: The Llm to monitor for tokens
        :param invocation_context: The InvocationContext
        :param journal: The OriginatingJournal which through which this
                    will send token count AGENT messages
        :param origin: The origin that will be applied to all messages.
        """
        self.llm: BaseLanguageModel = llm
        self.invocation_context: InvocationContext = invocation_context
        self.journal: Journal = journal
        self.origin: List[Dict[str, Any]] = origin
        self.debug: bool = False

    async def count_tokens(self, awaitable: Awaitable, max_execution_seconds: float = None) -> Any:
        """
        Counts the tokens (if possible) from what happens inside the awaitable
        within a separate context.  If tokens are counted, they are added to
        the InvocationContext's request_reporting and sent over the message queue
        via the journal

        Recall awaitables are a full async method call with args.  That is, where you would expect to
                baz = await myinstance.foo(bar)
        you instead do
                baz = await token_counter.count_tokens(myinstance.foo(bar)).

        :param awaitable: The awaitable whose tokens we wish to count.
        :param max_execution_seconds: The maximum amount of time to execute the awaitable.
                        If None, the awaitable is executed to completion.
        :return: Whatever the awaitable would return
        """

        retval: Any = None
        llm_factory: ContextTypeLlmFactory = self.invocation_context.get_llm_factory()
        llm_infos: Dict[str, Any] = llm_factory.llm_infos
        # Take a time stamp so we measure another thing people care about - latency.
        start_time: float = time()

        # Attempt to count tokens/costs while invoking the agent.
        # The means by which this happens is on a per-LLM basis, so get the right hook
        # given the LLM we've got.
        callback: AsyncCallbackHandler = None
        # Record origin information in our own context var so we can associate
        # with the langchain callback context vars more easily.
        origin_str: str = Origination.get_full_name_from_origin(self.origin)
        ORIGIN_INFO.set(origin_str)

        # Use the context manager to count tokens as per
        #   https://python.langchain.com/docs/how_to/llm_token_usage_tracking/#using-callbacks
        #
        # Caveats:
        # * In using this context manager approach, any tool that is called
        #   also has its token counts contributing to its callers for better or worse.
        # * As of 2/21/25, it seems that tool-calling agents (branch nodes) are not
        #   registering their tokens correctly. Not sure if this is a bug in langchain
        #   or there is something we are not doing in that scenario that we should be.
        # * As of 8/21/25, placing the journaling callback in the invoke config instead of llm
        #   appears to change the context manager’s behavior. The returned tokens from callback
        #   are now limited to the calling agent only, and no longer include those
        #   from downstream (chained) agents. However, `models_token_dict` is added
        #   to the `LlmTokenCallbackHandler` to collect token stats of each model call.
        with get_llm_token_callback(llm_infos) as callback:
            # Create a new context for different ContextVar values
            # and use the create_task() to run within that context.
            new_context: Context = copy_context()
            task: Task = new_context.run(self.create_task, awaitable)
            try:
                retval = await wait_for(task, max_execution_seconds)
            except AsyncTimeout:
                # Per docs for wait_for(), the task is already cancelled.
                retval = None

        # Figure out how much time our agent took.
        end_time: float = time()
        time_taken_in_seconds: float = end_time - start_time

        await self.report(callback, time_taken_in_seconds)

        return retval

    def create_task(self, awaitable: Awaitable) -> Task:
        """
        Riffed from:
        https://stackoverflow.com/questions/78659844/async-version-of-context-run-for-context-vars-in-python-asyncio
        """
        executor: AsyncioExecutor = self.invocation_context.get_asyncio_executor()
        origin_str: str = ORIGIN_INFO.get()
        task: Task = executor.create_task(awaitable, origin_str)

        if self.debug:
            # Print to be sure we have a different callback object.
            oai_call = llm_token_callback_var.get()
            print(f"origin is {origin_str} callback var is {id(oai_call)}")

        return task

    async def report(self, callback: AsyncCallbackHandler, time_taken_in_seconds: float):
        """
        Report on the token accounting results of the callback

        :param callback: An AsyncCallbackHandler or BaseCallbackHandle instance that contains token counting information
        :param time_taken_in_seconds: The amount of time the awaitable took in count_tokens()
        """

        # Accumulate what we learned about tokens to request reporting.
        # For now we just overwrite the one key because we know
        # the last one out will be the front man, and as of 2/21/25 his stats
        # are cumulative.  At some point we might want a finer-grained breakdown
        # that perhaps contributes to a service/er-wide periodic token stats breakdown
        # of some kind.  For now, get something going.
        #
        # Update (8/21/25):
        # Placing the journaling callback in the invoke config instead of llm changes the context
        # manager’s behavior. The returned tokens from the callback are now limited
        # to the calling agent only, not downstream (chained) agents.
        # Instead, `models` have been added to `request_reporting["token_accounting"]` to collect per-model stats
        # which are combined into network token stats.
        # Since the frontman is always the last to finish, by the time it exits,
        # `request_reporting["token_accounting"]` is complete and ready to report.
        request_reporting: Dict[str, Any] = self.invocation_context.get_request_reporting()
        token_accounting: Dict[str, Any] = request_reporting.get("token_accounting", {})
        models_token_dict: Dict[str, Any] = \
            self.merge_dicts(token_accounting.get("models", {}), callback.models_token_dict)
        network_token_dict: Dict[str, Any] = self.sum_all_tokens(models_token_dict, time_taken_in_seconds)
        # Provide sligtly different "caveats" for the network token accounting.
        network_token_dict["caveats"] = [
            "External agent token usage is not included.",
            "Token counts are approximate and estimated using tiktoken.",
            "time_taken_in_seconds includes overhead from Langchain and Neuro SAN"
        ]
        request_reporting["token_accounting"] = \
            {**network_token_dict, "models": models_token_dict}

        # Token counting results are collected in the callback.
        # Create a token counting dictionary for each agent
        agent_token_dict: Dict[str, Any] = self._generate_agent_token_dict(callback, time_taken_in_seconds)
        if self.journal is not None:
            # We actually have a token dictionary to report, so go there.
            agent_message = AgentMessage(structure=agent_token_dict)
            await self.journal.write_message(agent_message)
            # For frontman (origin with no ".") write both network token dict and model token dict
            if "." not in ORIGIN_INFO.get():
                network_token_message = AgentMessage(structure=request_reporting["token_accounting"])
                await self.journal.write_message(network_token_message)

    def _generate_agent_token_dict(
            self,
            callback: Union[AsyncCallbackHandler, BaseCallbackHandler],
            time_taken_in_seconds: float,
    ) -> Dict[str, Any]:
        """
        Generate the token counting dictionary for journals

        :param callback: An AsyncCallbackHandler or BaseCallbackHandler instance that contains
                            token counting information
        :param time_taken_in_seconds: The amount of time the awaitable took in count_tokens()
        :param agent_name: Name of the agent responsible for the token dictionary
        :return: Formatted token dictionary
        """

        # Organize the token dict for each agent to be the same format
        agent_token_dict = {
            "total_tokens": callback.total_tokens,
            "prompt_tokens": callback.prompt_tokens,
            "completion_tokens": callback.completion_tokens,
            "successful_requests": callback.successful_requests,
            "total_cost": callback.total_cost,
            "time_taken_in_seconds": time_taken_in_seconds,
            "caveats": [
                "Token usage is tracked at the agent level.",
                "Token counts are approximate and estimated using tiktoken.",
                "time_taken_in_seconds includes overhead from Langchain and Neuro SAN"
            ]
        }

        return agent_token_dict

    def sum_all_tokens(self, token_dict: Dict[str, Any], time_value: float) -> Dict[str, Any]:

        """
        Sum all token metrics across providers and models, **excluding time**.
        :param token_dict: Models token dict to aggregate into network stats
        :param time_value: Time taken for frontman to finish
        :return: Token stats of the entire network, either cumulative or single iteration
        """
        aggregated: Dict[str, Any] = {}
        for models in token_dict.values():
            for model_stats in models.values():
                for metric, value in model_stats.items():
                    if metric != "time_taken_in_seconds":
                        aggregated[metric] = aggregated.get(metric, 0) + value

        aggregated["time_taken_in_seconds"] = time_value

        return aggregated

    def merge_dicts(self, dict_1, dict_2):
        """
        Recursively merge two dictionaries.

        If both dictionaries contain the same key:
        - If the corresponding values are dictionaries, they are merged recursively.
        - Otherwise, the values are assumed to be numeric and are summed.

        Keys that exist only in one dictionary are carried over unchanged.

        :param dict_1: The base dictionary.
        :param dict_2: The dictionary whose values will be merged into `dict_1`.
        :return: A new dictionary containing the merged result.
        """
        result: Dict[str, Any] = dict(dict_1)  # start with dict_1
        for key, value in dict_2.items():
            if key in result:
                if isinstance(result[key], dict) and isinstance(value, dict):
                    # recursively merge nested dicts
                    result[key] = self.merge_dicts(result[key], value)
                else:
                    # assume values are numbers, sum them
                    result[key] += value
            else:
                result[key] = value
        return result




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.messages.traced_message import TracedMessage


class InterceptingJournal(Journal):
    """
    A Journal implementation that intercepts messages for a specific origin en route
    to another wrapped Journal.
    """

    def __init__(self, wrapped_journal: Journal, origin: List[Dict[str, Any]]):
        """
        Constructor

        :param wrapped_journal: The journal to forward messages to
        :param origin: The origin dictionary to match for intercepts
        """
        self.wrapped_journal: Journal = wrapped_journal
        self.origin: List[Dict[str, Any]] = origin
        self.messages: List[BaseMessage] = []

    async def write_message(self, message: BaseMessage, origin: List[Dict[str, Any]]):
        """
        Writes a BaseMessage entry into the journal
        :param message: The BaseMessage instance to write to the journal
        :param origin: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        # Let the wrapped guy do what he's gunna do
        await self.wrapped_journal.write_message(message, origin)
        self.write_unwrapped_message(message, origin)

    def write_unwrapped_message(self, message: BaseMessage, origin: List[Dict[str, Any]]):
        """
        Write a message to this journal without going through the wrapped journal.

        :param message: The BaseMessage instance to write to the journal
        :param origin: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """

        # Only consider messages that match the same origin as what we care about.
        if origin == self.origin:

            new_message: BaseMessage = message

            # When messages are TracedMessages, capture a version that
            # has all their fields translated to the additional_kwargs
            # for better display in tracing/observability applications like
            # LangSmith.
            if isinstance(message, TracedMessage):
                new_message = message.__class__(trace_source=message)

            self.messages.append(new_message)

    def get_messages(self) -> List[BaseMessage]:
        """
        :return: the intercepted messages
        """
        return self.messages




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple

import asyncio
import contextlib
import json
import tornado

from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.service.generic.async_agent_service import AsyncAgentService
from neuro_san.service.generic.async_agent_service_provider import AsyncAgentServiceProvider
from neuro_san.service.interfaces.agent_authorizer import AgentAuthorizer
from neuro_san.service.mcp.util.mcp_errors_util import McpErrorsUtil
from neuro_san.service.mcp.util.mcp_request_util import McpRequestUtil
from neuro_san.service.mcp.validation.tool_request_validator import ToolRequestValidator
from neuro_san.service.http.logging.http_logger import HttpLogger


class McpToolsProcessor:
    """
    Class implementing "tools"-related MCP requests.
    Overall MCP documentation can be found here:
    https://modelcontextprotocol.io/specification/2025-06-18/server/tools
    """

    def __init__(self,
                 logger: HttpLogger,
                 network_storage_dict: AgentNetworkStorage,
                 agent_policy: AgentAuthorizer,
                 tool_request_validator: ToolRequestValidator):
        self.logger: HttpLogger = logger
        self.network_storage_dict: AgentNetworkStorage = network_storage_dict
        self.agent_policy: AgentAuthorizer = agent_policy
        self.tool_request_validator: ToolRequestValidator = tool_request_validator

    async def list_tools(self, request_id, metadata: Dict[str, Any]) -> Dict[str, Any]:
        """
        List available MCP tools.
        :param request_id: MCP request id;
        :param metadata: http-level request metadata;
        :return: json dictionary with tools list in MCP format
        """
        # See which agents the user has access to per authorization policy
        authorized_agents: List[str] = await self.agent_policy.list_agents(metadata)

        public_storage: AgentNetworkStorage = self.network_storage_dict.get("public")
        tools_description: List[Dict[str, Any]] = []
        for agent_name in public_storage.get_agent_names():

            # Skip agents that are not authorized
            if agent_name not in authorized_agents:
                continue

            provider: AgentNetworkProvider = public_storage.get_agent_network_provider(agent_name)
            if provider is not None:
                agent_network: AgentNetwork = provider.get_agent_network()
                if agent_network.is_mcp_tool():
                    tool_dict: Dict[str, Any] = await self._get_tool_description(agent_name, metadata)
                    tools_description.append(tool_dict)
        return {
            "jsonrpc": "2.0",
            "id": McpRequestUtil.safe_request_id(request_id),
            "result": {
                "tools": tools_description
            }
        }

    # pylint: disable=too-many-return-statements
    async def call_tool(self, request_id, metadata: Dict[str, Any],
                        tool_name: str,
                        prompt: Dict[str, Any],
                        chat_context: Dict[str, Any],
                        chat_filter: Dict[str, Any],
                        sly_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Call MCP tool, which executes neuro-san agent chat request.
        :param request_id: MCP request id;
        :param metadata: http-level request metadata;
        :param tool_name: tool name;
        :param prompt: input prompt as a JSON structure;
        :param chat_context: chat context JSON structure, could be None;
        :param chat_filter: chat filter type JSON structure, could be None;
        :param sly_data: arbitrary JSON dictionary containing sly_data, could be None;
        :return: json dictionary with tool response in MCP format;
                 or json dictionary with error message in MCP format.
        """
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-arguments
        # pylint: disable=too-many-positional-arguments

        is_authorized: bool = False
        service_provider: AsyncAgentServiceProvider = None
        is_authorized, service_provider = await self.agent_policy.allow_agent(tool_name, metadata)

        if service_provider is None:
            # No such tool is found:
            return McpErrorsUtil.get_tool_error(request_id, f"Tool not found: {tool_name}")

        if not is_authorized:
            return McpErrorsUtil.get_tool_error(request_id, f"Tool not authorized: {tool_name}")

        service: AsyncAgentService = service_provider.get_service()
        if not service.is_mcp_tool():
            # Service is not allowed to be called as MCP tool:
            return McpErrorsUtil.get_tool_error(request_id, f"Service not available as MCP tool: {tool_name}")

        tool_timeout_seconds: float = service.get_request_timeout_seconds()
        if tool_timeout_seconds <= 0.0:
            # For asyncio.timeout(), None means no timeout:
            tool_timeout_seconds = None

        input_request: Dict[str, Any] = self._get_chat_input_request(prompt, chat_context, chat_filter, sly_data)
        response_text: str = ""
        response_structure: Dict[str, Any] = None
        try:
            async with asyncio.timeout(tool_timeout_seconds):
                result_generator = service.streaming_chat(input_request, metadata)
                async for result_dict in result_generator:
                    partial_response, structure_data = await self._extract_tool_response_part(result_dict)
                    if partial_response is not None:
                        response_text = response_text + partial_response
                    if structure_data is not None:
                        response_structure = structure_data

        except (asyncio.CancelledError, tornado.iostream.StreamClosedError):
            self.logger.info(metadata, "Tool execution %s cancelled/stream closed.", tool_name)
            return McpErrorsUtil.get_tool_error(request_id, f"Stream closed for tool {tool_name}")

        except asyncio.TimeoutError:
            self.logger.info(metadata,
                             "Chat tool timeout for %s in %f seconds.",
                             tool_name, tool_timeout_seconds)
            return McpErrorsUtil.get_tool_error(request_id, f"Timeout for tool {tool_name}")

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.logger.error(metadata, "Tool %s execution failed: %s", tool_name, str(exc))
            return McpErrorsUtil.get_tool_error(request_id, f"Failed to execute tool {tool_name}")

        finally:
            # We are done with the response stream,
            # ensure generator is closed properly in any case:
            if result_generator is not None:
                with contextlib.suppress(Exception):
                    # It is possible we will call .aclose() twice
                    # on our result_generator - it is allowed and has no effect.
                    await result_generator.aclose()

        # Return tool call result:
        call_result: Dict[str, Any] =\
            await self.build_tool_call_result(request_id, response_text, response_structure)
        return call_result

    async def build_tool_call_result(
            self,
            request_id,
            result_text: str,
            result_structure: Dict[str, Any]) -> Dict[str, Any]:
        """
        Build MCP tool call result dictionary from given text and structure parts.
        :param request_id: MCP request id;
        :param result_text: tool call result text part;
        :param result_structure: tool call result structure part;
        :return: json dictionary with tool call result in MCP format;
        """

        call_result: Dict[str, Any] = {
            "jsonrpc": "2.0",
            "id": McpRequestUtil.safe_request_id(request_id),
            "result": {
                "content": [
                    {
                        "type": "text",
                        "text": ""  # to be filled later
                    }
                ],
                "isError": False
            }
        }
        # Construct actual tool call result:
        if result_structure is not None:
            # "structuredContent" is MCP standard key for content with additional structure.
            call_result["result"]["structuredContent"] = result_structure
            # For backward compatibility, also add text version of structure:
            structure_data: Dict[str, Any] = result_structure.get("structure", None)
            if structure_data is not None:
                structure_str: str = f"```json\n{json.dumps(structure_data, indent=2)}\n```"
                result_text = result_text + structure_str
        call_result["result"]["content"][0]["text"] = McpRequestUtil.safe_message(result_text)
        return call_result

    async def _get_tool_description(self, agent_name: str, metadata: Dict[str, Any]) -> Dict[str, Any]:

        is_authorized: bool = False
        service_provider: AsyncAgentServiceProvider = None
        is_authorized, service_provider = await self.agent_policy.allow_agent(agent_name, metadata)

        if service_provider is None or not is_authorized:
            return None

        service: AsyncAgentService = service_provider.get_service()
        function_dict: Dict[str, Any] = await service.function({}, metadata)
        tool_description: str = function_dict.get("function", {}).get("description", "")
        return {
            "name": agent_name,
            "description": tool_description,
            "inputSchema": self.tool_request_validator.get_request_schema()
        }

    async def _extract_tool_response_part(
            self, response_dict: Dict[str, Any]) -> Tuple[str, Dict[str, Any]]:
        """
        Extract the tool response part from the given streaming chat response dictionary.
        :param response_dict: streaming chat response dictionary;
        :return: tuple of 2 values:
            text part as string or None,
            structure part as dictionary or None
        """
        response_part_dict: Dict[str, Any] = response_dict.get("response", {})
        response_type: str = response_part_dict.get("type", "")
        if response_type == "AGENT_FRAMEWORK":
            text: str = response_part_dict.get("text", None)
            # For final response, there could be chat_context structured data we need to return:
            structure_data: Dict[str, Any] = self.construct_mcp_structed_content(response_part_dict)
            return text, structure_data
        return None, None

    def construct_mcp_structed_content(self, response_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Construct MCP structured content dictionary from the given streaming chat response dictionary.
        :param response_dict: streaming chat response dictionary;
        :return: structured content dictionary or None
        """
        # We are looking for 2 possible parts of structured content:
        # "structure" - that was optionally generated by an agent;
        # and "chat_context" - that allows the context of a continuing conversation to be reconstructed by a client.
        # If both are missing, we return None.
        structure_data: Dict[str, Any] = response_dict.get("structure", None)
        chat_context_data: Dict[str, Any] = response_dict.get("chat_context", None)
        if structure_data is None and chat_context_data is None:
            return None
        result: Dict[str, Any] = {}
        if structure_data is not None:
            result["structure"] = structure_data
        if chat_context_data is not None:
            result["chat_context"] = chat_context_data
        return result

    def _get_chat_input_request(self,
                                user_message: Dict[str, Any],
                                chat_context: Dict[str, Any],
                                chat_filter: Dict[str, Any],
                                sly_data: Dict[str, Any]) -> Dict[str, Any]:
        """
        Construct a Python dictionary expected by "streaming_chat" service API call.
        :param user_message: user message JSON structure;
        :param chat_context: chat context JSON structure, could be None;
        :param chat_filter: chat filter type JSON structure, could be None;
        :param sly_data: arbitrary JSON dictionary containing sly_data, could be None;
        :return: "streaming_chat" service API input dictionary
        """
        request_dict: Dict[str, Any] = {
            "user_message": user_message
        }
        if chat_filter is not None:
            request_dict["chat_filter"] = chat_filter
        if chat_context is not None:
            request_dict["chat_context"] = chat_context
        if sly_data is not None:
            request_dict["sly_data"] = sly_data
        return request_dict
